import os
import sys
import logging
from typing import List, Tuple, Union, Callable
import functools
import math

import numpy as np
from scipy import sparse

import torch
import torch.nn as nn
import torch.nn.functional as F

import skorch
import skorch.utils
import dance.utils.loss as loss_functions
from torch.utils.data import DataLoader

REDUCE_LR_ON_PLATEAU_PARAMS = {
    "mode": "min",
    "factor": 0.1,
    "patience": 10,
    "min_lr": 1e-6,
}

def recursive_to_device(t, device="cpu"):
    """Recursively transfer t to the given device"""
    if isinstance(t, tuple) or isinstance(t, list):
        return tuple(recursive_to_device(x, device=device) for x in t)
    return t.to(device)

class Exp(nn.Module):
    """Applies torch.exp, clamped to improve stability during training"""

    def __init__(self, minimum=1e-5, maximum=1e6):
        """Values taken from DCA"""
        super(Exp, self).__init__()
        self.min_value = minimum
        self.max_value = maximum

    def forward(self, input):
        return torch.clamp(
            torch.exp(input),
            min=self.min_value,
            max=self.max_value,
        )

class ClippedSoftplus(nn.Module):
    def __init__(self, beta=1, threshold=20, minimum=1e-4, maximum=1e3):
        super(ClippedSoftplus, self).__init__()
        self.beta = beta
        self.threshold = threshold
        self.min_value = minimum
        self.max_value = maximum

    def forward(self, input):
        return torch.clamp(
            F.softplus(input, self.beta, self.threshold),
            min=self.min_value,
            max=self.max_value,
        )

    def extra_repr(self):
        return "beta={}, threshold={}, min={}, max={}".format(
            self.beta,
            self.threshold,
            self.min_value,
            self.max_value,
        )

class DeepCountAutoencoder(nn.Module):
    """Replicate of DCA"""

    def __init__(
        self,
        input_dim: int,
        inter_dim: int = 64,
        bottle_dim: int = 32,
        output_dim: int = None,
        mode: str = "zinb",
    ):
        super(DeepCountAutoencoder, self).__init__()
        torch.manual_seed(1234)  # Fixed initialization
        assert mode in ["zinb", "nb", "poisson"], f"Unrecognized mode: {mode}"
        self.mode = mode

        self.encoder = nn.Linear(input_dim, inter_dim)
        nn.init.xavier_uniform_(self.encoder.weight)  # Aka Glorot initialization
        self.bn1 = nn.BatchNorm1d(num_features=inter_dim)

        self.bottleneck = nn.Linear(inter_dim, bottle_dim)
        nn.init.xavier_uniform_(self.bottleneck.weight)
        self.bn2 = nn.BatchNorm1d(num_features=bottle_dim)

        self.decoder = nn.Linear(bottle_dim, inter_dim)
        nn.init.xavier_uniform_(self.decoder.weight)
        self.bn3 = nn.BatchNorm1d(num_features=inter_dim)

        output_dim = output_dim if output_dim is not None else input_dim
        self.mean = nn.Linear(
            inter_dim, output_dim
        )  # These parameters are all per-gene
        nn.init.xavier_uniform_(self.mean.weight)
        if "nb" in self.mode:
            self.disp = nn.Linear(inter_dim, output_dim)
            nn.init.xavier_uniform_(self.disp.weight)
        if self.mode == "zinb":
            self.dropout = nn.Linear(inter_dim, output_dim)
            nn.init.xavier_uniform_(self.dropout.weight)

        self.forward = self.forward_with_decode  # Default to returning denoised

    def encode(self, x):
        """Given input, return bottlenecked latent representation"""
        x = F.relu(self.bn1(self.encoder(x)))
        x = F.relu(self.bn2(self.bottleneck(x)))
        return x

    def decode(self, x, size_factors):
        """Given latent representation, output 3 output layers (mean, dispersion, dropout)"""
        x = F.relu(self.bn3(self.decoder(x)))
        mu = torch.clamp(torch.exp(self.mean(x)), min=1e-5, max=1e6)  # Mean
        sf_scaled = size_factors.view(-1, 1).repeat(1, mu.shape[1])
        mu_scaled = mu * sf_scaled  # Elementwise multiplication

        if "nb" in self.mode:
            theta = torch.clamp(
                F.softplus(self.disp(x)), min=1e-4, max=1e3
            )  # Dispersion
        else:  # Is poisson
            return mu_scaled

        if self.mode == "zinb":
            pi = torch.sigmoid(self.dropout(x))  # Dropout
            return mu_scaled, theta, pi
        else:
            return mu_scaled, theta

    def forward_with_decode(self, x, size_factors):
        """
        Return the parameters mu, theta, and pi
        The denoised matrix is generated by replacing the original count values with
        the mean of the negative binomial component as predicted in the output layer.
        This matrix represents the denoised and library size normalized expression matrix,
        the final output of the method.
        """
        return self.decode(self.encode(x), size_factors)

    def forward_no_decode(self, x, _size_factors):
        """
        Return the hidden representation instead of decoding it as well
        Useful for probing the latent dimension using pytorch_eval
        Use by setting model.forward = model.forward_no_decode
        """
        return self.encode(x)


class Encoder(nn.Module):
    def __init__(self, num_inputs: int, num_units=32, activation=nn.PReLU):
        super().__init__()
        self.num_inputs = num_inputs
        self.num_units = num_units

        self.encode1 = nn.Linear(self.num_inputs, 64)
        nn.init.xavier_uniform_(self.encode1.weight)
        self.bn1 = nn.BatchNorm1d(64)
        self.act1 = activation()

        self.encode2 = nn.Linear(64, self.num_units)
        nn.init.xavier_uniform_(self.encode2.weight)
        self.bn2 = nn.BatchNorm1d(num_units)
        self.act2 = activation()

    def forward(self, x):
        x = self.act1(self.bn1(self.encode1(x)))
        x = self.act2(self.bn2(self.encode2(x)))
        return x


class Decoder(nn.Module):
    def __init__(
        self,
        num_outputs: int,
        num_units: int = 32,
        intermediate_dim: int = 64,
        activation=nn.PReLU,
        final_activation=None,
    ):
        super().__init__()
        self.num_outputs = num_outputs
        self.num_units = num_units

        self.decode1 = nn.Linear(self.num_units, intermediate_dim)
        nn.init.xavier_uniform_(self.decode1.weight)
        self.bn1 = nn.BatchNorm1d(intermediate_dim)
        self.act1 = activation()

        self.decode21 = nn.Linear(intermediate_dim, self.num_outputs)
        nn.init.xavier_uniform_(self.decode21.weight)
        self.decode22 = nn.Linear(intermediate_dim, self.num_outputs)
        nn.init.xavier_uniform_(self.decode22.weight)
        self.decode23 = nn.Linear(intermediate_dim, self.num_outputs)
        nn.init.xavier_uniform_(self.decode23.weight)

        self.final_activations = nn.ModuleDict()
        if final_activation is not None:
            if isinstance(final_activation, list) or isinstance(
                final_activation, tuple
            ):
                assert len(final_activation) <= 3
                for i, act in enumerate(final_activation):
                    if act is None:
                        continue
                    self.final_activations[f"act{i+1}"] = act
            elif isinstance(final_activation, nn.Module):
                self.final_activations["act1"] = final_activation
            else:
                raise ValueError(
                    f"Unrecognized type for final_activation: {type(final_activation)}"
                )

    def forward(self, x, size_factors=None):
        """include size factor here because we may want to scale the output by that"""
        x = self.act1(self.bn1(self.decode1(x)))

        retval1 = self.decode21(x)  # This is invariably the counts
        if "act1" in self.final_activations.keys():
            retval1 = self.final_activations["act1"](retval1)
        if size_factors is not None:
            sf_scaled = size_factors.view(-1, 1).repeat(1, retval1.shape[1])
            retval1 = retval1 * sf_scaled  # Elementwise multiplication

        retval2 = self.decode22(x)
        if "act2" in self.final_activations.keys():
            retval2 = self.final_activations["act2"](retval2)

        retval3 = self.decode23(x)
        if "act3" in self.final_activations.keys():
            retval3 = self.final_activations["act3"](retval3)

        return retval1, retval2, retval3


class ChromEncoder(nn.Module):
    """
    Consumes multiple inputs (i.e. one feature vector for each chromosome)
    After processing everything to be the same dimensionality, concatenate
    to form a single latent dimension
    """

    def __init__(
        self, num_inputs: List[int], latent_dim: int = 32, activation=nn.PReLU
    ):
        super(ChromEncoder, self).__init__()
        self.num_inputs = num_inputs
        self.act = activation

        self.initial_modules = nn.ModuleList()
        for n in self.num_inputs:
            assert isinstance(n, int)
            layer1 = nn.Linear(n, 32)
            nn.init.xavier_uniform_(layer1.weight)
            bn1 = nn.BatchNorm1d(32)
            act1 = self.act()
            layer2 = nn.Linear(32, 16)
            nn.init.xavier_uniform_(layer2.weight)
            bn2 = nn.BatchNorm1d(16)
            act2 = self.act()
            self.initial_modules.append(
                nn.ModuleList([layer1, bn1, act1, layer2, bn2, act2])
            )

        self.encode2 = nn.Linear(16 * len(self.num_inputs), latent_dim)
        nn.init.xavier_uniform_(self.encode2.weight)
        self.bn2 = nn.BatchNorm1d(latent_dim)
        self.act2 = self.act()

    def forward(self, x):
        assert len(x) == len(
            self.num_inputs
        ), f"Expected {len(self.num_inputs)} inputs but got {len(x)}"
        enc_chroms = []
        for init_mod, chrom_input in zip(self.initial_modules, x):
            for f in init_mod:
                chrom_input = f(chrom_input)
            enc_chroms.append(chrom_input)
        enc1 = torch.cat(
            enc_chroms, dim=1
        )  # Concatenate along the feature dim not batch dim
        enc2 = self.act2(self.bn2(self.encode2(enc1)))
        return enc2


class ChromDecoder(nn.Module):
    """
    Network that is per-chromosome aware, but does not does not output
    per-chromsome values, instead concatenating them into a single vector
    """

    def __init__(
        self,
        num_outputs: List[int],  # Per-chromosome list of output sizes
        latent_dim: int = 32,
        activation=nn.PReLU,
        final_activations=[Exp(), ClippedSoftplus()],
    ):
        super().__init__()
        self.num_outputs = num_outputs
        self.latent_dim = latent_dim

        self.decode1 = nn.Linear(self.latent_dim, len(self.num_outputs) * 16)
        nn.init.xavier_uniform_(self.decode1.weight)
        self.bn1 = nn.BatchNorm1d(len(self.num_outputs) * 16)
        self.act1 = activation()

        self.final_activations = nn.ModuleDict()
        if final_activations is not None:
            if isinstance(final_activations, list) or isinstance(
                final_activations, tuple
            ):
                assert len(final_activations) <= 3
                for i, act in enumerate(final_activations):
                    if act is None:
                        continue
                    self.final_activations[f"act{i+1}"] = act
            elif isinstance(final_activations, nn.Module):
                self.final_activations["act1"] = final_activations
            else:
                raise ValueError(
                    f"Unrecognized type for final_activation: {type(final_activations)}"
                )
        logging.info(
            f"ChromDecoder with {len(self.final_activations)} output activations"
        )

        self.final_decoders = nn.ModuleList()  # List[List[Module]]
        for n in self.num_outputs:
            layer0 = nn.Linear(16, 32)
            nn.init.xavier_uniform_(layer0.weight)
            bn0 = nn.BatchNorm1d(32)
            act0 = activation()
            # l = [layer0, bn0, act0]
            # for _i in range(len(self.final_activations)):
            #     fc_layer = nn.Linear(32, n)
            #     nn.init.xavier_uniform_(fc_layer.weight)
            #     l.append(fc_layer)
            # self.final_decoders.append(nn.ModuleList(l))
            layer1 = nn.Linear(32, n)
            nn.init.xavier_uniform_(layer1.weight)
            layer2 = nn.Linear(32, n)
            nn.init.xavier_uniform_(layer2.weight)
            layer3 = nn.Linear(32, n)
            nn.init.xavier_uniform_(layer3.weight)
            self.final_decoders.append(
                nn.ModuleList([layer0, bn0, act0, layer1, layer2, layer3])
            )

    def forward(self, x):
        x = self.act1(self.bn1(self.decode1(x)))
        # This is the reverse operation of cat
        x_chunked = torch.chunk(x, chunks=len(self.num_outputs), dim=1)

        retval1, retval2, retval3 = [], [], []
        for chunk, processors in zip(x_chunked, self.final_decoders):
            # Each processor is a list of 3 different decoders
            # decode1, bn1, act1, *output_decoders = processors
            decode1, bn1, act1, decode21, decode22, decode23 = processors
            chunk = act1(bn1(decode1(chunk)))
            temp1 = decode21(chunk)
            temp2 = decode22(chunk)
            temp3 = decode23(chunk)

            if "act1" in self.final_activations.keys():
                # temp1 = output_decoders[0](chunk)
                temp1 = self.final_activations["act1"](temp1)
                # retval1.append(temp1)
            if "act2" in self.final_activations.keys():
                # temp2 = output_decoders[1](chunk)
                temp2 = self.final_activations["act2"](temp2)
                # retval2.append(temp2)
            if "act3" in self.final_activations.keys():
                # temp3 = output_decoders[2](chunk)
                temp3 = self.final_activations["act3"](temp3)
                # retval3.append(temp3)
            retval1.append(temp1)
            retval2.append(temp2)
            retval3.append(temp3)
        retval1 = torch.cat(retval1, dim=1)
        retval2 = torch.cat(retval2, dim=1)
        retval3 = torch.cat(retval3, dim=1)
        return retval1, retval2, retval3

        # retval = []
        # if retval1:
        #     retval.append(torch.cat(retval1, dim=1))
        # if retval2:
        #     retval.append(torch.cat(retval2, dim=1))
        # if retval3:
        #     retval.append(torch.cat(retval3, dim=1))
        # return tuple(retval)


class AutoEncoder(nn.Module):
    """Vanilla autoencoder"""

    def __init__(
        self,
        num_inputs: int,
        num_units: int = 32,
        num_outputs: int = None,
        activation=nn.PReLU,
        final_activation=None,
        seed=8947,
        output_encoded: bool = True,
    ):
        super().__init__()
        torch.manual_seed(seed)
        self.output_encoded = output_encoded
        self.num_inputs = num_inputs
        self.num_outputs = self.num_inputs if num_outputs is None else num_outputs
        self.num_units = num_units

        self.encoder = Encoder(
            self.num_inputs, num_units=self.num_units, activation=activation
        )
        self.decoder = Decoder(
            self.num_outputs,
            num_units=self.num_units,
            activation=activation,
            final_activation=final_activation,
        )

    def forward(self, X):
        encoded = self.encoder(X)
        decoded = self.decoder(encoded)[
            0
        ]  # Drop the second output that corresponds to addtl parameters
        if self.output_encoded:
            return decoded, encoded
        return decoded


class PairedAutoEncoder(nn.Module):
    """
    Paired autoencoder
    Supports cross-domain prediction by naively combining/swapping
    encoder and decoder
    """

    def __init__(self, model1, model2):
        super().__init__()
        self.model1 = model1
        self.model2 = model2

    def forward(self, x):
        """x is expected to be a tuple of two inputs"""
        x1, x2 = x
        y1 = self.model1(x1)
        y2 = self.model2(x2)
        return (y1, y2)

    def translate_1_to_2(self, encoded1):
        """Given data from domain 1 output domain 2"""
        output2 = self.model2.from_encoded(encoded1)
        return output2

    def translate_2_to_1(self, encoded2):
        """Given data fromd omain 2 output domain 1"""
        output1 = self.model1.from_encoded(encoded2)
        return output1


class SplicedAutoEncoder(nn.Module):
    """
    Spliced Autoencoder - where we have 4 parts (2 encoders, 2 decoders) that are all combined
    This does not work when you have chromsome split features
    """

    def __init__(
        self,
        input_dim1: int,
        input_dim2: int,
        hidden_dim: int = 16,
        final_activations1: list = [Exp(), nn.Softplus()],
        final_activations2 = [Exp(), nn.Softplus(), nn.Sigmoid()],
        flat_mode: bool = False,  # Controls if we have to re-split inputs
        seed=182822,
    ):
        super().__init__()
        torch.manual_seed(seed)

        self.flat_mode = flat_mode
        self.input_dim1 = input_dim1
        self.input_dim2 = input_dim2
        self.num_outputs1 = (
            len(final_activations1)
            if isinstance(final_activations1, (list, set, tuple))
            else 1
        )
        self.num_outputs2 = (
            len(final_activations2)
            if isinstance(final_activations2, (list, set, tuple))
            else 1
        )

        self.encoder1 = Encoder(num_inputs=input_dim1, num_units=hidden_dim)
        self.encoder2 = Encoder(num_inputs=input_dim2, num_units=hidden_dim)

        self.decoder1 = Decoder(
            num_outputs=input_dim1,
            num_units=hidden_dim,
            final_activation=final_activations1,
        )
        self.decoder2 = Decoder(
            num_outputs=input_dim2,
            num_units=hidden_dim,
            final_activation=final_activations2,
        )

    def split_catted_input(self, x):
        """
        Split catted input data to expected sizes
        """
        return torch.split(x, [self.input_dim1, self.input_dim2], dim=-1)

    def _combine_output_and_encoded(self, decoded, encoded, num_outputs: int):
        """
        Combines the output and encoded in a single output
        """
        if num_outputs > 1:
            retval = *decoded, encoded
        else:
            if isinstance(decoded, tuple):
                decoded = decoded[0]
            retval = decoded, encoded
        assert isinstance(retval, (list, tuple))
        assert isinstance(
            retval[0], (torch.TensorType, torch.Tensor)
        ), f"Expected tensor but got {type(retval[0])}"
        return retval

    def forward_single(
        self, x, size_factors=None, in_domain: int = 1, out_domain: int = 1
    ):
        """
        Return output of a single domain combination
        """
        assert in_domain in [1, 2] and out_domain in [1, 2]

        encoder = self.encoder1 if in_domain == 1 else self.encoder2
        decoder = self.decoder1 if out_domain == 1 else self.decoder2
        num_non_latent_out = self.num_outputs1 if out_domain == 1 else self.num_outputs2

        encoded = encoder(x)
        decoded = decoder(encoded)
        return self._combine_output_and_encoded(decoded, encoded, num_non_latent_out)

    def forward(self, x, size_factors=None, mode: Union[None, Tuple[int, int]] = None):
        if self.flat_mode:
            x = self.split_catted_input(x)
        assert isinstance(x, (tuple, list))
        assert len(x) == 2, "There should be two inputs to spliced autoencoder"
        encoded1 = self.encoder1(x[0])
        encoded2 = self.encoder2(x[1])

        decoded11 = self.decoder1(encoded1)
        retval11 = self._combine_output_and_encoded(
            decoded11, encoded1, self.num_outputs1
        )
        decoded12 = self.decoder2(encoded1)
        retval12 = self._combine_output_and_encoded(
            decoded12, encoded1, self.num_outputs2
        )
        decoded22 = self.decoder2(encoded2)
        retval22 = self._combine_output_and_encoded(
            decoded22, encoded2, self.num_outputs2
        )
        decoded21 = self.decoder1(encoded2)
        retval21 = self._combine_output_and_encoded(
            decoded21, encoded2, self.num_outputs1
        )

        if mode is None:
            return retval11, retval12, retval21, retval22
        retval_dict = {
            (1, 1): retval11,
            (1, 2): retval12,
            (2, 1): retval21,
            (2, 2): retval22,
        }
        if mode not in retval_dict:
            raise ValueError(f"Invalid mode code: {mode}")
        return retval_dict[mode]


class NaiveSplicedAutoEncoder(SplicedAutoEncoder):
    """
    Naive "spliced" autoencoder that does not use shared branches and instead simply
    trains four separate models
    """

    def __init__(
        self,
        input_dim1: int,
        input_dim2: int,
        hidden_dim: int = 16,
        final_activations1: Union[Callable, List[Callable]] = [
            Exp(),
            ClippedSoftplus(),
        ],
        final_activations2: Union[Callable, List[Callable]] = nn.Sigmoid(),
        flat_mode: bool = True,  # Controls if we have to re-split inputs
        seed: int = 182822,
    ):
        nn.Module.__init__(self)
        torch.manual_seed(seed)

        self.flat_mode = flat_mode
        self.input_dim1 = input_dim1
        self.input_dim2 = input_dim2
        self.num_outputs1 = (
            len(final_activations1)
            if isinstance(final_activations1, (list, set, tuple))
            else 1
        )
        self.num_outputs2 = (
            len(final_activations2)
            if isinstance(final_activations2, (list, set, tuple))
            else 1
        )

        self.model11 = nn.ModuleList(
            [
                Encoder(num_inputs=input_dim1, num_units=hidden_dim),
                Decoder(
                    num_outputs=input_dim1,
                    num_units=hidden_dim,
                    final_activation=final_activations1,
                ),
            ]
        )
        self.model12 = nn.ModuleList(
            [
                Encoder(num_inputs=input_dim1, num_units=hidden_dim),
                ChromDecoder(
                    num_outputs=input_dim2,
                    latent_dim=hidden_dim,
                    final_activations=final_activations2,
                ),
            ]
        )
        self.model22 = nn.ModuleList(
            [
                ChromEncoder(num_inputs=input_dim2, latent_dim=hidden_dim),
                ChromDecoder(
                    num_outputs=input_dim2,
                    latent_dim=hidden_dim,
                    final_activations=final_activations2,
                ),
            ]
        )
        self.model21 = nn.ModuleList(
            [
                ChromEncoder(num_inputs=input_dim2, latent_dim=hidden_dim),
                Decoder(
                    num_outputs=input_dim1,
                    num_units=hidden_dim,
                    final_activation=final_activations1,
                ),
            ]
        )
        # Each of these has exactly two items
        self.modlist_dict = nn.ModuleDict(
            {
                "11": self.model11,
                "12": self.model12,
                "22": self.model22,
                "21": self.model21,
            }
        )

    def split_catted_input(self, x):
        """Split the input into chunks that goes to each input to model"""
        a, b = torch.split(x, [self.input_dim1, sum(self.input_dim2)], dim=-1)
        return (a, torch.split(b, self.input_dim2, dim=-1))

    def forward_single(
        self, x, size_factors=None, in_domain: int = 1, out_domain: int = 1
    ):
        """
        Return output of a single domain combination
        """
        num_non_latent_out = self.num_outputs1 if out_domain == 1 else self.num_outputs2
        modlist = self.modlist_dict[str(in_domain) + str(out_domain)]

        encoded = modlist[0](x)
        decoded = modlist[1](encoded)

        if num_non_latent_out > 1:
            retval = *decoded, encoded
        else:
            if isinstance(decoded, tuple):
                decoded = decoded[0]
            retval = decoded, encoded

        assert isinstance(retval, tuple)
        assert isinstance(retval[0], (torch.TensorType, torch.Tensor))
        return retval

    def forward(self, x, size_factors=None, mode: Union[None, Tuple[int, int]] = None):
        if self.flat_mode:
            x = self.split_catted_input(x)
        assert isinstance(x, (tuple, list))
        assert len(x) == 2, "There should be two inputs to spliced autoencoder"
        retval11 = self.forward_single(
            x[0], size_factors=size_factors, in_domain=1, out_domain=1
        )
        retval12 = self.forward_single(
            x[0], size_factors=size_factors, in_domain=1, out_domain=2
        )
        retval21 = self.forward_single(
            x[1], size_factors=size_factors, in_domain=2, out_domain=1
        )
        retval22 = self.forward_single(
            x[1], size_factors=size_factors, in_domain=2, out_domain=2
        )

        if mode is None:
            return retval11, retval12, retval21, retval22
        retval_dict = {
            (1, 1): retval11,
            (1, 2): retval12,
            (2, 1): retval21,
            (2, 2): retval22,
        }
        if mode not in retval_dict:
            raise ValueError(f"Invalid mode code: {mode}")
        return retval_dict[mode]


class AssymSplicedAutoEncoder(SplicedAutoEncoder):
    """
    Assymmetric spliced autoencoder where branch 2 is a chrom AE
    """

    def __init__(
        self,
        input_dim1: int,
        input_dim2: List[int],
        hidden_dim: int = 16,
        final_activations1: list = [Exp(), ClippedSoftplus()],
        final_activations2=nn.Sigmoid(),
        flat_mode: bool = True,  # Controls if we have to re-split inputs
        seed: int = 182822,
    ):
        # https://stackoverflow.com/questions/9575409/calling-parent-class-init-with-multiple-inheritance-whats-the-right-way
        nn.Module.__init__(self)
        torch.manual_seed(seed)

        self.flat_mode = flat_mode
        self.input_dim1 = input_dim1
        self.input_dim2 = input_dim2
        self.num_outputs1 = (
            len(final_activations1)
            if isinstance(final_activations1, (list, set, tuple))
            else 1
        )
        self.num_outputs2 = (
            len(final_activations2)
            if isinstance(final_activations2, (list, set, tuple))
            else 1
        )

        self.encoder1 = Encoder(num_inputs=input_dim1, num_units=hidden_dim)
        self.encoder2 = ChromEncoder(num_inputs=input_dim2, latent_dim=hidden_dim)

        self.decoder1 = Decoder(
            num_outputs=input_dim1,
            num_units=hidden_dim,
            final_activation=final_activations1,
        )
        self.decoder2 = ChromDecoder(
            num_outputs=input_dim2,
            latent_dim=hidden_dim,
            final_activations=final_activations2,
        )

    def split_catted_input(self, x):
        """Split the input into chunks that goes to each input to model"""
        a, b = torch.split(x, [self.input_dim1, sum(self.input_dim2)], dim=-1)
        return (a, torch.split(b, self.input_dim2, dim=-1))

class PairedAutoEncoderSkorchNet(skorch.NeuralNet):
    def forward_iter(self, X, training=False, device="cpu"):
        """Subclassed to work with tuples"""
        dataset = self.get_dataset(X)
        iterator = self.get_iterator(dataset, training=training)
        for i, data in enumerate(iterator):
            Xi = skorch.dataset.unpack_data(data)[0]
            with torch.set_grad_enabled(training):
                self.module_.train(training)
                yp = self.infer(Xi)

            if isinstance(yp, tuple):
                yield recursive_to_device(yp)  # <- modification here
            else:
                yield yp.to(device)

    def predict_proba(self, x):
        """Subclassed so calling predict produces a tuple of outputs"""
        y_probas1, y_probas2 = [], []
        for yp in self.forward_iter(x, training=False):
            assert isinstance(yp, tuple)
            yp1 = yp[0][0]
            yp2 = yp[1][0]
            y_probas1.append(skorch.utils.to_numpy(yp1))
            y_probas2.append(skorch.utils.to_numpy(yp2))
        y_proba1 = np.concatenate(y_probas1, 0)
        y_proba2 = np.concatenate(y_probas2, 0)
        return y_proba1, y_proba2

    def get_encoded_layer(self, x):
        """Get the encoded representation as a TUPLE of two elements"""
        encoded1, encoded2 = [], []
        for out1, out2, *_other in self.forward_iter(x, training=False):
            encoded1.append(out1[-1])
            encoded2.append(out2[-1])
        return np.concatenate(encoded1, axis=0), np.concatenate(encoded2, axis=0)

    def translate_1_to_2(self, x):
        enc1, enc2 = self.get_encoded_layer(x)
        device = next(self.module_.parameters()).device
        enc1_torch = torch.from_numpy(enc1).to(device)
        return self.module_.translate_1_to_2(enc1_torch)[0].detach().cpu().numpy()

    def translate_2_to_1(self, x):
        enc1, enc2 = self.get_encoded_layer(x)
        device = next(self.module_.parameters()).device
        enc2_torch = torch.from_numpy(enc2).to(device)
        return self.module_.translate_2_to_1(enc2_torch)[0].detach().cpu().numpy()


class SplicedAutoEncoderSkorchNet(PairedAutoEncoderSkorchNet):
    """
    Skorch wrapper for the SplicedAutoEncoder above.
    Mostly here to take care of how we calculate loss
    """

    def predict_proba(self, x):
        """
        Subclassed so that calling predict produces a tuple of 4 outputs
        """
        y_probas1, y_probas2, y_probas3, y_probas4 = [], [], [], []
        for yp in self.forward_iter(x, training=False):
            assert isinstance(yp, tuple)
            yp1 = yp[0][0]
            yp2 = yp[1][0]
            yp3 = yp[2][0]
            yp4 = yp[3][0]
            y_probas1.append(skorch.utils.to_numpy(yp1))
            y_probas2.append(skorch.utils.to_numpy(yp2))
            y_probas3.append(skorch.utils.to_numpy(yp3))
            y_probas4.append(skorch.utils.to_numpy(yp4))
        y_proba1 = np.concatenate(y_probas1)
        y_proba2 = np.concatenate(y_probas2)
        y_proba3 = np.concatenate(y_probas3)
        y_proba4 = np.concatenate(y_probas4)
        # Order: 1to1, 1to2, 2to1, 2to2
        return y_proba1, y_proba2, y_proba3, y_proba4

    def get_encoded_layer(self, x):
        """Get the encoded representation as a TUPLE of two elements"""
        encoded1, encoded2 = [], []
        for out11, out12, out21, out22 in self.forward_iter(x, training=False):
            encoded1.append(out11[-1])
            encoded2.append(out22[-1])
        return np.concatenate(encoded1, axis=0), np.concatenate(encoded2, axis=0)

    def translate_1_to_1(self, x) -> sparse.csr_matrix:
        retval = [
            sparse.csr_matrix(skorch.utils.to_numpy(yp[0][0]))
            for yp in self.forward_iter(x, training=False)
        ]
        return sparse.vstack(retval)

    def translate_1_to_2(self, x) -> sparse.csr_matrix:
        retval = [
            sparse.csr_matrix(skorch.utils.to_numpy(yp[1][0]))
            for yp in self.forward_iter(x, training=False)
        ]
        return sparse.vstack(retval)

    def translate_2_to_1(self, x) -> sparse.csr_matrix:
        retval = [
            sparse.csr_matrix(skorch.utils.to_numpy(yp[2][0]))
            for yp in self.forward_iter(x, training=False)
        ]
        return sparse.vstack(retval)

    def translate_2_to_2(self, x) -> sparse.csr_matrix:
        retval = [
            sparse.csr_matrix(skorch.utils.to_numpy(yp[3][0]))
            for yp in self.forward_iter(x, training=False)
        ]
        return sparse.vstack(retval)

    def score(self, true, pred):
        return self.get_loss(pred, true)


class PairedInvertibleAutoEncoderSkorchNet(PairedAutoEncoderSkorchNet):
    def translate_1_to_2(self, x):
        enc1, enc2 = self.get_encoded_layer(x)
        device = next(self.module_.parameters()).device
        enc1_torch = torch.from_numpy(enc1).to(device)
        return self.module_.translate_1_to_2(enc1_torch)[0].detach().cpu().numpy()

    def translate_2_to_1(self, x):
        enc1, enc2 = self.get_encoded_layer(x)
        device = next(self.module_.parameters()).device
        enc2_torch = torch.from_numpy(enc2).to(device)
        return self.module_.translate_2_to_1(enc2_torch)[0].detach().cpu().numpy()

# BabelWrapper = SplicedAutoEncoderSkorchNet
class BabelWrapper:
    """Babel class.

    Parameters
    ----------
    args : argparse.Namespace
        A Namespace object that contains arguments of Babel. For details of parameters in parser args, please refer to link (parser help document).
    dataset : dance.datasets.multimodality.ModalityPredictionDataset
        Modality prediction dataset.
    """

    def __init__(self, args, dataset):
        self.args = args
        model_class = (
            NaiveSplicedAutoEncoder
            if args.naive
            else AssymSplicedAutoEncoder
        )
        self.model = model_class(hidden_dim=args.hidden,
            input_dim1=dataset.modalities[0].shape[1],
            input_dim2=[dataset.modalities[1].shape[1]],
            final_activations1=nn.ReLU(),
            final_activations2=nn.ReLU(),
            flat_mode=True,
            seed=args.rnd_seed,
            ).to(args.device)

    def to(self, device):
        """Performs device conversion.

        Parameters
        ----------
        device : str
            Target device.

        Returns
        -------
        self : BabelWrapper
            Converted model.
        """
        self.model.to(device)

    def score(self, test_mod1, test_mod2):
        """Score function to get score of prediction.

        Parameters
        ----------
        test_mod1 : torch.Tensor
            Input modality features.
        test_mod2 : torch.Tensor
            Target modality features.

        Returns
        -------
        score : float
            RMSE loss of prediction.

        """
        mse = nn.MSELoss()
        self.model.eval()

        with torch.no_grad():
            pred = self.predict(test_mod1)
            score = math.sqrt(mse(pred, test_mod2.to(self.args.device)))
            return score

    def predict(self, test_mod1):
        """Predict function to get prediction of target modality features.

        Parameters
        ----------
        test_mod1 : torch.Tensor
            Input modality features.

        Returns
        -------
        pred : torch.Tensor
            Prediction of target modality features.

        """
        self.model.eval()

        with torch.no_grad():
            if self.args.naive:
                emb = self.model.model12[0](test_mod1.to(self.args.device))
                pred = self.model.model12[1](emb)[0]
            else:
                emb = self.model.encoder1(test_mod1.to(self.args.device))
                pred = self.model.decoder2(emb)[0]
            return pred

    def fit(self, train_dual, valid_dual, max_epochs=500):
        """fit function for training.

        Parameters
        ----------
        train_dual : dance.utils.PairedDataset
            Training dataset.
        valid_dual : dance.utils.PairedDataset
            Validation dataset.
        max_epochs : int optional
            Maximum number of training epochs, by default to be 500.

        Returns
        -------
        None.
        """

        #TODO: add callbacks
        # callbacks = [
        #                 skorch.callbacks.EarlyStopping(patience=self.args.earlystop),
        #                 skorch.callbacks.GradientNormClipping(gradient_clip_value=5),
        #             ],
        # train_dual = sc_dual_train_dataset
        # valid_dual = sc_dual_valid_dataset
        criterion = loss_functions.QuadLoss(loss1=loss_functions.RMSELoss, loss2=loss_functions.RMSELoss,
                                            loss2_weight=self.args.lossweight)
        device = self.args.device

        optimizer = torch.optim.Adam(self.model.parameters(), lr=self.args.lr)
        train_loader = DataLoader(dataset=train_dual,
                                  batch_size=self.args.batchsize,
                                  shuffle=True,
                                  num_workers=1, )
        valid_loader = DataLoader(dataset=valid_dual,
                                  batch_size=len(valid_dual),
                                  shuffle=False,
                                  num_workers=1, )

        val = []
        for i in range(max_epochs):
            self.model.train()
            total_loss = 0
            for idx, data in enumerate(train_loader):
                logits = self.model(data[0].to(device))
                loss = criterion(logits, data[1].to(device))
                total_loss += loss.item()

                optimizer.zero_grad()
                loss.backward()

                nn.utils.clip_grad_norm_(self.model.parameters(), 5)
                optimizer.step()

            mse = nn.MSELoss()
            self.model.eval()

            with torch.no_grad():
                loss = 0
                for _, data in enumerate(valid_loader):
                    logits = self.model(data[0].to(device))
                    loss += mse(logits[1][0], data[1][:, -logits[1][0].shape[1]:].to(device)).item()
            val.append(math.sqrt(loss/len(valid_loader)))
            print('epoch: ', i+1)
            print('training (sum of 4 losses):', total_loss/len(train_loader))
            print('validation (prediction loss):', val[-1])

            if min(val) == val[-1]:
                torch.save(self.model.state_dict(), f'{self.args.outdir}/BABEL_best_{self.args.rnd_seed}.pth')
            if i>self.args.earlystop and min(val) != min(val[-self.args.earlystop:]):
                print('Early stopped.')
                self.model.load_state_dict(torch.load(f'{self.args.outdir}/BABEL_best_{self.args.rnd_seed}.pth'))
                break



"""Reimplementation of the scDeepSort cell-type annotation method.

Reference
---------
Shao, Xin, et al. "scDeepSort: a pre-trained cell-type annotation method for single-cell transcriptomics using deep
learning with a weighted graph neural network." Nucleic acids research 49.21 (2021): e122-e122.

"""
import os
import time
from pathlib import Path

import dgl.function as fn
import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from dgl.dataloading import DataLoader, NeighborSampler

DEBUG = os.environ.get("DANCE_DEBUG")


class AdaptiveSAGE(nn.Module):
    """The AdaptiveSAGE graph convolution layer.

    Similar to SAGE convolution with mean aggregation, but with additional flexibility that adaptively assigns
    importance to gene-cell interactions, as well as gene and cell self-loops. In particular, each gene will be
    associated with an importance score `beta` that are used as aggregation weights by the cell nodes. Additionally,
    there are two `alpha` parameters indicating how much each cell or gene should retain its previous representations.

    Note
    ----
    In practice, `alpha` and `beta` are stored in a unified tensor called `alpha`. The first #gene elements of this
    tensor are the `beta` values and the last two elements are the actual `alpha` values.

    """

    def __init__(self, dim_in, dim_out, alpha, dropout_layer, act_layer, norm_layer):
        """Initialize the AdaptiveSAGE convolution layer.

        Parameters
        ----------
        dim_in : int
            Input feature dimensions.
        dim_out : int
            output feature dimensinos.
        alpha : Tensor
            Shared learnable parameters containing gene-cell interaction strengths and those for the cell and gene
            self-loops.
        dropout_layer : torch.nn
            Dropout layer.
        act_layer : torch.nn
            Activation layer.
        norm_layer : torch.nn
            Normalization layer.

        """
        super().__init__()

        self.alpha = alpha
        self.gene_num = len(alpha) - 2

        self.layers = nn.ModuleList()
        self.layers.append(dropout_layer)
        self.layers.append(nn.Linear(dim_in, dim_out))
        nn.init.xavier_uniform_(self.layers[-1].weight, gain=nn.init.calculate_gain("relu"))
        self.layers.append(act_layer)
        self.layers.append(norm_layer)

    def message_func(self, edges):
        """Message update function.

        Reweight messages based on 1) the shared learnable interaction strengths and 2) the underlying edgeweights of
        the graph. In particular, for 1), gene-cell interaction (undirectional) will be weighted by the gene specific
        `beta` value, and the cell and gene self-interactions will be weighted based on the corresponding `alpha`
        values.

        """
        number_of_edges = edges.src["h"].shape[0]
        indices = np.expand_dims(np.array([self.gene_num + 1] * number_of_edges, dtype=np.int32), axis=1)
        src_id, dst_id = edges.src["id"].cpu().numpy(), edges.dst["id"].cpu().numpy()
        indices = np.where((src_id >= 0) & (dst_id < 0), src_id, indices)  # gene->cell
        indices = np.where((dst_id >= 0) & (src_id < 0), dst_id, indices)  # cell->gene
        indices = np.where((dst_id >= 0) & (src_id >= 0), self.gene_num, indices)  # gene-gene
        if DEBUG:
            print(
                f"{((src_id >= 0) & (dst_id < 0)).sum():>10,} (geen->cell), "
                f"{((src_id < 0) & (dst_id >= 0)).sum():>10,} (cell->gene), "
                f"{((src_id >= 0) & (dst_id >= 0)).sum():>10,} (self-gene), "
                f"{((src_id < 0) & (dst_id < 0)).sum():>10,} (self-cell), ",
            )
        h = edges.src["h"] * self.alpha[indices.squeeze()]
        return {"m": h * edges.data["weight"]}

    def forward(self, block, h):
        with block.local_scope():
            h_src = h
            h_dst = h[:block.number_of_dst_nodes()]
            block.srcdata["h"] = h_src
            block.dstdata["h"] = h_dst
            block.update_all(self.message_func, fn.mean("m", "neigh"))

            z = block.dstdata["h"]
            for layer in self.layers:
                z = layer(z)

            return z


class GNN(nn.Module):
    """The scDeepSort GNN model.

    Message passing between cell and genes to learn cell representations for cell-type annotations. The graph contains
    both cell and gene nodes. The gene features are initialized as PCA embeddings from the (normalized) scRNA-seq
    samples matrix, and the cell features are computed as the weighted aggregation of the gene features according to
    each cell's gene expressions.

    """

    def __init__(self, dim_in, dim_out, dim_hid, n_layers, gene_num, activation=None, norm=None, dropout=0.):
        """Initialize the scDeepSort GNN model.

        Parameters
        ----------
        dim_in : int
            Input dimension (PCA embeddings dimension).
        dim_out : int
            Output dimension (number of unique cell labels).
        n_layers : int
            Number of convolution layers.
        gene_num : int
            Number of genes.
        dropout : float
            Dropout ratio.
        activation : torch.nn, optional
            Activation layer.
        norm : torch.nn, optional
            Normalization layer.

        """
        super().__init__()

        self.n_layers = n_layers
        self.gene_num = gene_num

        # [gene_num] is alpha of gene-gene self loop, [gene_num+1] is alpha of cell-cell self loop, the rest are betas
        self.alpha = nn.Parameter(torch.tensor([1] * (self.gene_num + 2), dtype=torch.float32).unsqueeze(-1))

        dropout_layer = nn.Dropout(p=dropout) if dropout > 0 else nn.Identity()
        act_layer = activation or nn.Identity()
        norm_layer = norm or nn.Identity()

        self.layers = nn.ModuleList()
        for i in range(n_layers):
            input_dimension = dim_in if i == 0 else dim_hid
            self.layers.append(AdaptiveSAGE(input_dimension, dim_hid, self.alpha, dropout_layer, act_layer, norm_layer))

        self.linear = nn.Linear(dim_hid, dim_out)
        nn.init.xavier_uniform_(self.linear.weight, gain=nn.init.calculate_gain("relu"))

    def forward(self, blocks, x):
        assert len(blocks) == len(self.layers), f"Inonsistent layer info: {len(blocks)=} vs {len(self.layers)=}"
        for block, layer in zip(blocks, self.layers):
            x = layer(block, x)
        return self.linear(x)


class ScDeepSort:
    """The ScDeepSort cell-type annotation model.

    Parameters
    ----------
    params : argparse.Namespace
        A Namespace contains arguments of Scdeepsort. See parser documnetation for more details.

    """

    def __init__(self, params):
        self.params = params
        self.postfix = time.strftime("%d_%m_%Y") + "_" + time.strftime("%H:%M:%S")
        self.prj_path = Path(__file__).resolve().parents[4]
        # TODO: change the prefix from `example` to `saved_models`
        self.save_path = (
            self.prj_path
            / "example"
            / "single_modality"
            / "cell_type_annotation"
            / "pretrained"
            / self.params.species
            / "models"
        )

        if not self.save_path.exists():
            self.save_path.mkdir(parents=True)
        self.device = torch.device("cpu" if self.params.gpu == -1 else f"cuda:{params.gpu}")

        # Define the variables during training
        self.num_cells = None
        self.num_genes = None
        self.num_labels = None
        self.graph = None
        self.train_ids = None
        self.test_ids = None
        self.labels = None

        # Define the variables during prediction
        self.id2label = None
        self.test_dict = None

    def fit(self, num_cells, num_genes, num_labels, graph, train_ids, test_ids, labels):
        """Train scDeepsort model

        Parameters
        ----------
        num_cells : int
            The number of cells in the training set.
        num_genes : int
            The number of genes in the training set.
        num_labels : int
            The number of labels in the training set.
        graph : dgl.DGLGraph
            Training graph.
        train_ids : Tensor
            The training ids.
        test_ids : Tensor
            The testing ids.
        labels : Tensor
            Node (cell, gene) labels, -1 for genes.

        """
        self.num_cells = num_cells
        self.num_genes = num_genes
        self.num_labels = num_labels

        self.train_ids = train_ids.to(self.device)
        self.test_ids = test_ids.to(self.device)
        self.labels = labels.to(self.device)
        self.graph = graph.to(self.device)
        self.graph.ndata["label"] = self.labels

        self.model = GNN(self.params.dense_dim, self.num_labels, self.params.hidden_dim, self.params.n_layers,
                         self.num_genes, activation=nn.ReLU(), dropout=self.params.dropout).to(self.device)
        print(self.model)

        self.sampler = NeighborSampler(fanouts=[-1] * self.params.n_layers, edge_dir="in")
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.params.lr,
                                          weight_decay=self.params.weight_decay)
        self.loss_fn = nn.CrossEntropyLoss(reduction="sum")
        if self.params.num_neighbors == 0:
            self.num_neighbors = self.num_cells + self.num_genes
        else:
            self.num_neighbors = self.params.num_neighbors

        print(f"Train Number: {len(self.train_ids)}, Test Number: {len(self.test_ids)}")
        max_test_acc, _train_acc, _epoch = 0, 0, 0
        for epoch in range(self.params.n_epochs):
            loss = self.cal_loss()
            train_acc = self.evaluate(self.train_ids)[-1]
            test_correct, test_unsure, test_acc = self.evaluate(self.test_ids)
            if max_test_acc <= test_acc:
                final_test_correct_num = test_correct
                final_test_unsure_num = test_unsure
                _train_acc = train_acc
                _epoch = epoch
                max_test_acc = test_acc
                self.save_model()
            print(f">>>>Epoch {epoch:04d}: Train Acc {train_acc:.4f}, Loss {loss / len(self.train_ids):.4f}, "
                  f"Test correct {test_correct}, Test unsure {test_unsure}, Test Acc {test_acc:.4f}")

        print(f"---{self.params.species} {self.params.tissue} Best test result:---")
        print(f"Epoch {_epoch:04d}, Train Acc {_train_acc:.4f}, Test Correct Num {final_test_correct_num}, "
              f"Test Total Num {len(self.test_ids)}, Test Unsure Num {final_test_unsure_num}, "
              f"Test Acc {final_test_correct_num / len(self.test_ids):.4f}")

    def cal_loss(self):
        """Calculate loss.

        Returns
        -------
        float
            Loss function value.

        """
        self.model.train()
        total_loss = 0

        dataloader = DataLoader(graph=self.graph, indices=self.train_ids, graph_sampler=self.sampler,
                                batch_size=self.params.batch_size, shuffle=True)
        for _, _, blocks in dataloader:
            blocks = [b.to(self.device) for b in blocks]
            input_features = blocks[0].srcdata["features"]
            output_labels = blocks[-1].dstdata["label"]
            output_predictions = self.model(blocks, input_features)

            loss = self.loss_fn(output_predictions, output_labels)
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            total_loss += loss.item()

        return total_loss

    @torch.no_grad()
    def evaluate(self, ids):
        """Evaluate the trained scDeepsort model.

        Parameters
        ----------
        ids : Tensor
            A 1-D tensor containing node IDs to be evaluated on.

        Returns
        -------
        Tuple[int, int, float]
            The total number of correct prediction, the total number of unsure prediction, and the accuracy score.

        """
        self.model.eval()
        total_correct, total_unsure = 0, 0

        dataloader = DataLoader(graph=self.graph, indices=ids, graph_sampler=self.sampler,
                                batch_size=self.params.batch_size, shuffle=True)
        for _, _, blocks in dataloader:
            blocks = [b.to(self.device) for b in blocks]
            input_features = blocks[0].srcdata["features"]
            output_labels = blocks[-1].dstdata["label"]
            output_predictions = self.model(blocks, input_features)

            for pred, label in zip(output_predictions.cpu(), output_labels.cpu()):
                max_prob = pred.max().item()
                if max_prob < self.params.unsure_rate / self.num_labels:
                    total_unsure += 1
                elif pred.argmax().item() == label:
                    total_correct += 1

        return total_correct, total_unsure, total_correct / len(ids)

    def save_model(self):
        """Save the model at the save_path."""
        state = {"model": self.model.state_dict(), "optimizer": self.optimizer.state_dict()}
        torch.save(state, self.save_path / f"{self.params.species}-{self.params.tissue}.pt")

    def load_model(self):
        """Load the model from the model path."""
        filename = f"{self.params.species}-{self.params.tissue}.pt"
        model_path = self.prj_path / "pretrained" / self.params.species / "models" / filename
        state = torch.load(model_path, map_location=self.device)
        self.model.load_state_dict(state["model"])

    @torch.no_grad()
    def inference(self, num):
        """Perform inference on a test dataset.

        Parameters
        ----------
        num : int
            Test dataset number.

        Returns
        -------
        list
            Predicted labels.

        """
        self.model.eval()

        unsure_threshold = self.params.unsure_rate / self.num_labels

        graph = self.test_dict["graph"][num].to(self.device)
        test_indices = self.test_dict["nid"][num].to(self.device)
        new_logits = torch.zeros((graph.number_of_nodes(), self.num_labels))

        dataloader = DataLoader(graph=graph, indices=test_indices, graph_sampler=self.sampler,
                                batch_size=self.params.batch_size, shuffle=True)
        for _, output_nodes, blocks in dataloader:
            blocks = [b.to(torch.device(self.device)) for b in blocks]
            input_features = blocks[0].srcdata["features"]
            new_logits[output_nodes] = self.model(blocks, input_features).detach().cpu()

        new_logits = new_logits[self.test_dict["mask"][num]]
        new_logits = nn.functional.softmax(new_logits, dim=1).numpy()
        predict_label = []
        for pred in new_logits:
            pred_label = self.id2label[pred.argmax().item()]
            predict_label.append("unsure" if pred.max().item() < unsure_threshold else pred_label)
        return predict_label

    def predict(self, id2label, test_dict):
        """Perform prediction on all test datasets.

        Parameters
        ----------
        id2label : np.ndarray
            Id to label diction.
        test_dict : dict
            The test dictionary.

        Returns
        -------
        dict
            A dictionary where the keys are the test dataset IDs and the values are the corresponding predictions.

        """
        self.id2label = id2label
        self.test_dict = test_dict
        return {num: self.inference(num) for num in self.params.test_dataset}

    @torch.no_grad()
    def score(self, predicted_labels, true_labels):
        """Compute model performance on test datasets based on accuracy.

        Parameters
        ----------
        predicted_labels : dict
            A dictionary where the keys are test dataset IDs and the values are the predicted labels.
        true_labels : dict
            A dictionary where the keys are test dataset IDs and the values are the true labels of the cells. Each
            element, i.e., the label, can be either a specific value (e.g., string or intger) or a set of values,
            allowing multiple mappings.

        Returns
        -------
        dict
            A diction of correct prediction numbers, total samples, unsured prediction numbers, and accuracy.

        """
        output_score = {}
        for num in set(predicted_labels) & set(true_labels):
            total_num = len(predicted_labels[num])
            unsure_num = correct = 0
            for pred, true in zip(predicted_labels[num], true_labels[num]):
                if pred == "unsure":
                    unsure_num += 1
                elif pred == true or pred in true:  # either a single mapping or multiple mappings
                    correct += 1

            output_score[num] = {
                "Total number of predictions": total_num,
                "Number of correct predictions": correct,
                "Number of unsure predictions": unsure_num,
                "Accuracy": correct / total_num,
            }

        return output_score

    def save_pred(self, num, pred):
        """Save predictions for a particular test dataset.

        Parameters
        ----------
        num : int
            Test file number.
        pred : list np.array or dataframe
            Predicted labels.

        """
        label_map = pd.read_excel("./map/celltype2subtype.xlsx", sheet_name=self.params.species, header=0,
                                  names=["species", "old_type", "new_type", "new_subtype"])
        label_map = label_map.fillna("N/A", inplace=False)
        oldtype2newtype = {}
        oldtype2newsubtype = {}
        for _, old_type, new_type, new_subtype in label_map.itertuples(index=False):
            oldtype2newtype[old_type] = new_type
            oldtype2newsubtype[old_type] = new_subtype

        save_path = self.prj_path / self.params.save_dir
        if not save_path.exists():
            save_path.mkdir()
        if self.params.score:
            df = pd.DataFrame({
                "index": self.test_dict["origin_id"][num],
                "original label": self.test_dict["label"][num],
                "cell_type": [oldtype2newtype.get(p, p) for p in pred],
                "cell_subtype": [oldtype2newsubtype.get(p, p) for p in pred]
            })
        else:
            df = pd.DataFrame({
                "index": self.test_dict["origin_id"][num],
                "cell_type": [oldtype2newtype.get(p, p) for p in pred],
                "cell_subtype": [oldtype2newsubtype.get(p, p) for p in pred]
            })
        df.to_csv(save_path / (self.params.species + f"_{self.params.tissue}_{num}.csv"), index=False)
        print(f"output has been stored in {self.params.species}_{self.params.tissue}_{num}.csv")



import torch
import torch.nn as nn 
import torch.nn.functional as F
from torch.autograd import Variable
from torch.nn.parameter import Parameter
import torch.optim as optim
import math
import time
import numpy as np


class GraphConvolution(nn.Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """
    
    """GraphConvolution.

        Parameters
        ----------
        in_features : int 
            input dimension.
        out_features : int 
            output dimension.
        support : 
            support for graph convolution.    
        bias : boolean optional
            include bias term, default False.

        Returns
        -------
        None. 

        """
    
    def __init__(self, in_features, out_features, support, bias=False):
        super(GraphConvolution, self).__init__()
        self.support = support
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        #glorot
        init_range = np.sqrt(6.0 / (self.in_features + self.out_features))
        self.weight.data.uniform_(-init_range, init_range)
        # similar to kaiming normal/uniform
        stdv = 1. / math.sqrt(self.weight.size(1))
        #self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input, adj):
        """forward function.

        Parameters
        ----------
        input : 
            node features.
        adj : 
            adjacency matrix.

        Returns
        -------
        output : float
            output of graph convolution layer.

        """
        #convolution
        if input.is_sparse:
            #sparse input features
            support = torch.spmm(input, self.weight)
        else:
            support = torch.mm(input, self.weight)
        
        #adj should always be sparse
        #### add a ReLU or other activation!!
        output = torch.spmm(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'
    
    
class GCN(nn.Module):
    """
    dropout + GC + activation
    """
    def __init__(self, nfeat, nhid1, nout, bias=False, dropout=0., act=F.relu):
        super(GCN, self).__init__()
            
        self.gc1 = GraphConvolution(nfeat, nhid1, bias)
        self.gc2 = GraphConvolution(nhid1, nout, bias)
        self.dropout = dropout
        self.act = act
        

    def forward(self, x, adj):
        """forward function.

        Parameters
        ----------
        x : 
            node features.
        adj : 
            adjacency matrix.

        Returns
        -------
        output : float
            output of graph convolution network.

        """
        #self.nnz = x._nnz() if sparse_inputs else None
        
        #dropout + convolution
        x=dropout_layer(x, self.dropout)
        x = self.gc1(x, adj)
        x = self.act(x)
        x=dropout_layer(x, self.dropout)
        x=self.gc2(x, adj)
        
        return x
    

class DSTGLearner(object):
    """DSTGLearner.

        Parameters
        ----------
        nfeat : int 
            input dimension.
        nhid1 : int
            number of units in the hidden layer (graph convolution).
        nout : int 
            output dimension.
        bias : boolean optional
            include bias term, default False.
        dropout : float optional
            dropout rate, default 0.
        act : optional
            activation function, default torch functional relu.

        Returns
        -------
        None. 

        """
    def __init__(self, nfeat, nhid1, nout, bias=False, dropout=0., act=F.relu, device = torch.device("cuda" if torch.cuda.is_available() else "cpu")):
        super(DSTGLearner, self).__init__()
        self.device = device
        self.model = GCN(nfeat, nhid1, nout, bias, dropout, act).to(device)
            
    def fit(self, X, adj, labels, labels_mask=None, lr=0.005, max_epochs=100, weight_decay=0):
        """fit function for model training.

        Parameters
        ----------
        X : 
            node features.
        adj : 
            adjacency matrix.
        labels :
            labels of pseudo spots (and real spots if provided). 
        labels_mask : optional
            mask to indicate which samples to use for training.
        lr : float optional
            learning rate.
        max_epochs : int optional
            maximum number of epochs to train.
        weight_decay : float optional
            weight decay parameter for optimization (Adam). 
            
        Returns
        -------
        None.

        """
        X=X.to(self.device)
        adj=adj.to(self.device)
        labels=labels.to(self.device)
        labels_mask=labels_mask.to(self.device)
        
        #device = self.device
        model = self.model
        model.train()
        optimizer = optim.Adam(model.parameters(),lr=lr, weight_decay=weight_decay)
        # optimizer = RAdam(model.parameters(), lr=self.args.lr)
        
        #X=torch.FloatTensor(X)
        #adj=torch.FloatTensor(adj)
        
        for epoch in range(max_epochs):
            t = time.time()
            
            y_hat = model(X, adj)
            loss = masked_softmax_cross_entropy(y_hat, labels, labels_mask)
            
            if (epoch+1) % 5 == 0:
                print("Epoch:", '%04d' % (epoch + 1), "train_loss=",
                  "{:.5f}".format(loss),
                  "time=", "{:.5f}".format(time.time() - t))
            #    _ = model(X, adj)
            
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            
    def predict(self, X, adj):
        """prediction function.
        Parameters
        ----------
        X : 
            node features.
        adj : 
            adjacency matrix.
     
        Returns
        -------
        
        pred : torch tensor
            predictions of cell-type proportions.
        
        """
        X=X.to(self.device)
        adj=adj.to(self.device)
        fX=self.model(X,adj)
        pred = F.softmax(fX, dim=1)
        return pred
    
    
    def score(self, X, adj, labels, labels_mask=None):
        """score.

        Parameters
        ----------
        X : 
            node features.
        adj : 
            adjacency matrix.
        labels :
            pseudo spot labels (and real labels if provided).
        labels_mask : optional
            mask to indicate which samples to use in scoring.

        Returns
        -------
        loss : float
            cross entropy loss between predicted and true labels.

        """
        X=X.to(self.device)
        adj=adj.to(self.device)
        labels=labels.to(self.device)
        labels_mask=labels_mask.to(self.device)
        model = self.model
        model.eval()
        y_hat = model(X, adj)
        pred = F.softmax(y_hat, dim=1)
        loss = masked_softmax_cross_entropy(y_hat, labels, labels_mask)
        return loss

    
def dropout_layer(x, dropout):
    """dropout_layer.
    Parameters
    ----------
    x:  
        input to dropout layer.
    dropout : float
        dropout rate (between 0 and 1).

    Returns
    -------
    out : torch tensor
        dropout output.
    """
    if x.is_sparse:
        #sparse input features
        out = sparse_dropout(x, dropout)
        return out
    else:
        out = F.dropout(x, dropout)
        return out
        
        
def sparse_dropout(x, dropout):
    """sparse_dropout.
    Parameters
    ----------
    x:  
        input to dropout layer.
    dropout : float
        dropout rate (between 0 and 1).

    Returns
    -------
    out * (1. / (1-dropout)) : torch tensor
        dropout output.
    """
                        
    noise_shape = x._nnz()
    random_tensor = (1-dropout)+torch.rand(noise_shape).to(x.device)
    dropout_mask = torch.floor(random_tensor).type(torch.bool)
    i = x._indices()
    v = x._values()

    i = i[:, dropout_mask]
    v = v[dropout_mask]

    out = torch.sparse.FloatTensor(i, v, x.shape).to(x.device)
    return out * (1. / (1-dropout))


"""Softmax cross-entropy loss with masking."""
def masked_softmax_cross_entropy(preds, labels, mask):
    """masked_softmax_cross_entropy.
    Parameters
    ----------
    preds:  
        cell-type proportion predictions from dstg model.
    labels : 
        true cell-type proportion labels.
    mask : 
        mask to indicate which samples to use in computing loss.

    Returns
    -------
    loss : float
        cross entropy loss between true and predicted cell-type proportions (mean reduced).
    """
    if(mask is None):
        loss = F.cross_entropy(preds, labels, reduction='mean')
        return loss
    else: 
        loss = F.cross_entropy(preds, labels, reduction='none')
        mask = mask.type(torch.float32)
        mask /= torch.mean(mask)
        loss *= mask
        loss = torch.mean(loss)
        return loss


import os
import sys
import logging
import argparse
import numpy as np
import pandas as pd
import scipy.spatial
import scanpy as sc

import matplotlib.pyplot as plt

import torch
import torch.nn as nn
import torch.nn.functional as F
import random
from dance.utils import PairedDataset
import skorch.helper
import math

from dance.datasets.multimodality import ModalityPredictionDataset
from dance.modules.multi_modality.predict_modality.babel import BabelWrapper
from dance.utils import set_seed
import dance.utils.loss as loss_functions

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    OPTIMIZER_DICT = {
        "adam": torch.optim.Adam,
        "rmsprop": torch.optim.RMSprop,
    }
    rndseed = random.randint(0, 2147483647)
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--subtask', default='openproblems_bmmc_cite_phase2_rna')
    parser.add_argument('-device', '--device', default='cuda:3')
    parser.add_argument('-cpu', '--cpus', default = 1, type = int)
    parser.add_argument('-seed', '--rnd_seed', default = rndseed, type=int)
    parser.add_argument('-m', '--model_folder', default = './models')
    parser.add_argument("--outdir", "-o", default = './logs', help="Directory to output to")
    parser.add_argument("--lossweight", type=float, default=1., help="Relative loss weight",)
    parser.add_argument("--lr", "-l", type=float, default=0.01, help="Learning rate")
    parser.add_argument("--batchsize", "-b", type=int, default=64, help="Batch size")
    parser.add_argument("--hidden", type=int, default=64, help="Hidden dimensions")
    parser.add_argument("--earlystop", type=int, default=20, help="Early stopping after N epochs")
    parser.add_argument("--naive", "-n", action="store_true", help="Use a naive model instead of lego model",)
    parser.add_argument("--resume", action="store_true")
    parser.add_argument("--max_epochs", type=int, default=500)
    args = parser.parse_args()
    args.resume = True

    torch.set_num_threads(args.cpus)
    rndseed = args.rnd_seed
    set_seed(rndseed)
    dataset = ModalityPredictionDataset(args.subtask).load_data().preprocess('feature_selection')
    device = args.device
    os.makedirs(args.model_folder, exist_ok=True)
    os.makedirs(args.outdir, exist_ok=True)

    args.outdir = os.path.abspath(args.outdir)

    if not os.path.isdir(os.path.dirname(args.outdir)):
        os.makedirs(os.path.dirname(args.outdir))

    # Specify output log file
    logger = logging.getLogger()
    fh = logging.FileHandler(f"{args.outdir}/training_{args.subtask}_{args.rnd_seed}.log", "w")
    fh.setLevel(logging.INFO)
    logger.addHandler(fh)

    for arg in vars(args):
        logging.info(f"Parameter {arg}: {getattr(args, arg)}")

    idx = np.random.permutation(dataset.modalities[0].shape[0])
    train_idx = idx[:int(idx.shape[0]*0.85)]
    val_idx = idx[int(idx.shape[0]*0.85):]
    train_mod1 = torch.from_numpy(dataset.numpy_features(0)[train_idx]).float()
    train_mod2 = torch.from_numpy(dataset.numpy_features(1)[train_idx]).float()
    valid_mod1 = torch.from_numpy(dataset.numpy_features(0)[val_idx]).float()
    valid_mod2 = torch.from_numpy(dataset.numpy_features(1)[val_idx]).float()
    test_mod1 = torch.from_numpy(dataset.numpy_features(2)).float()
    test_mod2 = torch.from_numpy(dataset.numpy_features(3)).float()

    sc_dual_train_dataset = PairedDataset(train_mod1, train_mod2)
    sc_dual_valid_dataset = PairedDataset(valid_mod1, valid_mod2)

    model = BabelWrapper(args, dataset)
    model.fit(sc_dual_train_dataset, sc_dual_valid_dataset, args.max_epochs)
    print(model.predict(test_mod1))
    print(model.score(test_mod1, test_mod2))


    # # Instantiate and train model
    # model_class = (
    #     autoencoders.NaiveSplicedAutoEncoder
    #     if args.naive
    #     else autoencoders.AssymSplicedAutoEncoder
    # )
    #
    # spliced_net = autoencoders.BabelWrapper(
    #     module=model_class,
    #     module__hidden_dim=args.hidden,  # Based on hyperparam tuning
    #     module__input_dim1=dataset.modalities[0].shape[1],
    #     module__input_dim2=[dataset.modalities[1].shape[1]],
    #     module__final_activations1=[
    #                 autoencoders.Exp(),
    #                 autoencoders.ClippedSoftplus(),
    #             ],#nn.ReLU(),
    #     module__final_activations2=[
    #                 autoencoders.Exp(),
    #                 autoencoders.ClippedSoftplus(),
    #             ],#nn.ReLU(),
    #     module__flat_mode=True,
    #     module__seed=args.rnd_seed,
    #     lr=args.lr,  # Based on hyperparam tuning
    #     criterion=loss_functions.QuadLoss,
    #     criterion__loss1=loss_functions.NegativeBinomialLoss, #RMSELoss,
    #     criterion__loss2=loss_functions.NegativeBinomialLoss, #RMSELoss,  # handle output of encoded layer
    #     criterion__loss2_weight=1.5,#args.lossweight,  # numerically balance the two losses with different magnitudes
    #     criterion__record_history=True,
    #     optimizer=OPTIMIZER_DICT['adam'],
    #     iterator_train__shuffle=True,
    #     device=torch.device(args.device),
    #     batch_size=args.batchsize,  # Based on  hyperparam tuning
    #     max_epochs=500,
    #     callbacks=[
    #         skorch.callbacks.EarlyStopping(patience=args.earlystop),
    #         skorch.callbacks.LRScheduler(
    #             policy=torch.optim.lr_scheduler.ReduceLROnPlateau,
    #             **autoencoders.REDUCE_LR_ON_PLATEAU_PARAMS,
    #         ),
    #         skorch.callbacks.GradientNormClipping(gradient_clip_value=5),
    #         skorch.callbacks.Checkpoint(
    #             dirname=args.outdir, fn_prefix="net_", monitor="valid_loss_best",
    #         ),
    #     ],
    #     train_split=skorch.helper.predefined_split(sc_dual_valid_dataset),
    #     iterator_train__num_workers=8,
    #     iterator_valid__num_workers=8,
    # )
    #
    # if args.resume:
    #     # Load in the warm start parameters
    #     try:
    #         spliced_net.initialize()
    #         cp = skorch.callbacks.Checkpoint(dirname=args.outdir, fn_prefix="net_", monitor="valid_loss_best", )
    #         spliced_net.load_params(checkpoint=cp)
    #
    #         # Upon successfully finding correct hiden size, break out of loop
    #         print(f"Loaded model with hidden size {args.hidden}")
    #         spliced_net = spliced_net
    #     except RuntimeError as e:
    #         print(f"Failed to load with hidden size {args.hidden}")
    #     spliced_net.module_.eval()
    # else:
    #     spliced_net.fit(sc_dual_train_dataset, y=None)
    #     spliced_net.module_.eval()
    #
    # sc_rna_atac_train_preds = spliced_net.translate_1_to_2(sc_dual_train_dataset)
    # sc_rna_atac_valid_preds = spliced_net.translate_1_to_2(sc_dual_valid_dataset)
    # sc_rna_atac_test_preds = spliced_net.translate_1_to_2(sc_dual_test_dataset)
    #
    # def evaluate(pred, label):
    #     mse = nn.MSELoss()
    #     return math.sqrt(mse(torch.from_numpy(pred.todense()), label))
    #
    # print('train:', evaluate(sc_rna_atac_train_preds, train_mod2))
    # print('valid:', evaluate(sc_rna_atac_valid_preds, valid_mod2))
    # print('test:', evaluate(sc_rna_atac_test_preds, test_mod2))


#######################
##  Remake
#######################

#
# model = model_class(hidden_dim=args.hidden,
#             input_dim1=dataset.modalities[0].shape[1],
#             input_dim2=[dataset.modalities[1].shape[1]],
#             final_activations1=nn.ReLU(),
#             final_activations2=nn.ReLU(),
#             flat_mode=True,
#             seed=args.rnd_seed,
#             ).to(device)
#
# def train(
#         criterion=loss_functions.QuadLoss(loss1 = loss_functions.RMSELoss, loss2 = loss_functions.RMSELoss,
#                                           loss2_weight=args.lossweight),
#         max_epochs=500,
#         callbacks=[
#                 skorch.callbacks.EarlyStopping(patience=args.earlystop),
#                 skorch.callbacks.GradientNormClipping(gradient_clip_value=5),
#             ],
#         train_split=sc_dual_valid_dataset):
#
#     device = args.device
#
#     optimizer = torch.optim.Adam(model.parameters(), lr=args.lr)
#     train_loader = DataLoader(dataset=sc_dual_train_dataset,
#                               batch_size=args.batchsize,
#                               shuffle=True,
#                               num_workers=1, )
#     valid_loader = DataLoader(dataset=sc_dual_valid_dataset,
#                               batch_size=len(sc_dual_valid_dataset),
#                               shuffle=False,
#                               num_workers=1, )
#
#     for i in range(max_epochs):
#         model.train()
#         total_loss = 0
#         for idx, data in enumerate(train_loader):
#             logits = model(data[0].to(device))
#             loss = criterion(logits, data[1].to(device))
#             total_loss += loss.item()
#
#             optimizer.zero_grad()
#             loss.backward()
#             optimizer.step()
#
#         print('training (sum of 4 losses):', total_loss/len(train_loader))
#         print('validation (prediction loss):', evaluate(valid_loader))
#         test()
#
#
# def evaluate(valid_loader):
#     mse = nn.MSELoss()
#     model.eval()
#
#     with torch.no_grad():
#         loss = 0
#         for _, data in enumerate(valid_loader):
#             logits = model(data[0].to(device))
#             loss += mse(logits[1][0], data[1][:, -logits[1][0].shape[1]:].to(device)).item()
#         return math.sqrt(loss/len(valid_loader))
#
# def test():
#     mse = nn.MSELoss()
#     model.eval()
#
#     with torch.no_grad():
#         emb = model.model12[0](test_mod1.to(device))
#         pred = model.model12[1](emb)[0]
#         print(math.sqrt(mse(pred, test_mod2.to(device))))



# Copyright 2022 DSE lab.  All rights reserved.
from scipy import sparse
import torch
from networkx.algorithms import bipartite
import numpy as np
import scib
import scanpy as sc
from sklearn.neighbors import NearestNeighbors
from sklearn.cluster import KMeans
from sklearn.metrics.cluster import normalized_mutual_info_score
from sklearn.metrics import adjusted_rand_score
import random
import pylab as plt
import seaborn as sns
from scipy import sparse


def get_bipartite_matching_adjacency_matrix_mk3(raw_logits, threshold_quantile=0.995, copy=False):
    #getting rid of unpromising graph connections
    if copy:
        weights = raw_logits.copy()
    else:
        weights = raw_logits
    quantile_row = np.quantile(weights, threshold_quantile, axis=0, keepdims=True)
    quantile_col = np.quantile(weights, threshold_quantile, axis=1, keepdims=True)
    #quantile_minimum = np.minimum(quantile_row, quantile_col, out=quantile_row)
    mask_ = (weights<quantile_row)
    mask_ = np.logical_and(mask_, (weights<quantile_col), out=mask_)
    #weights[weights<quantile_minimum] = 0
    weights[mask_] = 0
    weights_sparse = sparse.csr_matrix(-weights)
    del(weights)
    graph = bipartite.matrix.from_biadjacency_matrix(weights_sparse)
    #explicitly combining top nodes in once component or networkx freaks tf out
    u = [n for n in graph.nodes if graph.nodes[n]['bipartite'] == 0]
    matches = bipartite.matching.minimum_weight_full_matching(graph, top_nodes=u)
    best_matches = np.array([matches[x]-len(u) for x in u])
    bipartite_matching_adjacency = np.zeros(raw_logits.shape)
    bipartite_matching_adjacency[np.arange(raw_logits.shape[0]), best_matches]=1
    return bipartite_matching_adjacency

def batch_separated_bipartite_matching(dataset, emb1, emb2):
    matrix = np.zeros((dataset.modalities[2].shape[0], dataset.modalities[2].shape[0]))
    for b in dataset.modalities[2].obs['batch'].unique():
        i0 = (dataset.modalities[2].obs['batch'] == b).values.nonzero()[0].tolist()
        j0 = (dataset.modalities[3].obs['batch'] == b).values.nonzero()[0].tolist()
        logits = torch.matmul(emb1[i0], emb2[j0].T)
        logits = torch.softmax(logits, -1) + torch.softmax(logits, 0)
        logits = logits.cpu().numpy()

        out1_2 = get_bipartite_matching_adjacency_matrix_mk3(logits, threshold_quantile=0.950)
        matrix[np.ix_(i0, j0)] = out1_2
    return matrix

def get_nmi(adata):
    # print('Preprocessing')
    sc.pp.neighbors(adata, use_rep='X_emb')
    # print('Clustering')
    scib.cl.opt_louvain(
        adata,
        label_key='cell_type',
        cluster_key='cluster',
        plot=False,
        inplace=True,
        force=True
    )
    # print('Compute score')
    score = scib.me.nmi(adata, group1='cluster', group2='cell_type')
    return score

def get_cell_type_ASW(adata):
    return scib.me.silhouette(adata, group_key='cell_type', embed='X_emb')

def get_cell_cycle_conservation(adata, adata_solution):
    recompute_cc = 'S_score' not in adata_solution.obs_keys() or \
            'G2M_score' not in adata_solution.obs_keys()
    organism = adata_solution.uns['organism']
    # print('Compute score')
    score = scib.me.cell_cycle(
        adata_pre=adata_solution,
        adata_post=adata,
        batch_key='batch',
        embed='X_emb',
        recompute_cc=recompute_cc,
        organism=organism
    )
    return score

def get_traj_conservation(adata, adata_solution):
    adt_atac_trajectory = 'pseudotime_order_ATAC' if 'pseudotime_order_ATAC' in adata_solution.obs else 'pseudotime_order_ADT'
    sc.pp.neighbors(adata, use_rep='X_emb')
    obs_keys = adata_solution.obs_keys()
    if 'pseudotime_order_GEX' in obs_keys:
        score_rna = scib.me.trajectory_conservation(
            adata_pre=adata_solution,
            adata_post=adata,
            label_key='cell_type',
            pseudotime_key='pseudotime_order_GEX'
        )
    else:
        score_rna = np.nan

    if adt_atac_trajectory in obs_keys:
        score_adt_atac = scib.me.trajectory_conservation(
            adata_pre=adata_solution,
            adata_post=adata,
            label_key='cell_type',
            pseudotime_key=adt_atac_trajectory
        )
    else:
        score_adt_atac = np.nan

    score_mean = (score_rna + score_adt_atac) / 2
    return score_mean

def get_batch_ASW(adata):
    score = scib.me.silhouette_batch(
        adata,
        batch_key='batch',
        group_key='cell_type',
        embed='X_emb',
        verbose=False
    )
    return score

def get_graph_connectivity(adata):
    sc.pp.neighbors(adata, use_rep='X_emb')
    # print('Compute score')
    score = scib.me.graph_connectivity(adata, label_key='cell_type')
    return score

def knn_accuracy(A, B, k, train_idx, test_idx, metric='l1'):
    nn = NearestNeighbors(k, metric=metric)
    nn.fit(A)
    transp_nearest_neighbor = nn.kneighbors(B, 1, return_distance=False)
    actual_nn = nn.kneighbors(A, k, return_distance=False)
    train_correct = 0
    test_correct = 0

    for i in range(len(transp_nearest_neighbor)):
        if transp_nearest_neighbor[i] not in actual_nn[i]:
            continue
        elif i in test_idx:
            test_correct += 1
        else:
            train_correct += 1

    return train_correct / len(train_idx), test_correct / len(test_idx)

def cluster_acc(y_true, y_pred):
    """
    Calculate clustering accuracy. Require scikit-learn installed
    # Arguments
        y: true labels, numpy.array with shape `(n_samples,)`
        y_pred: predicted labels, numpy.array with shape `(n_samples,)`
    # Return
        accuracy, in [0,1]
    """
    y_true = y_true.astype(np.int64)
    assert y_pred.size == y_true.size
    D = max(y_pred.max(), y_true.max()) + 1
    w = np.zeros((D, D), dtype=np.int64)
    for i in range(y_pred.size):
        w[y_pred[i], y_true[i]] += 1
    from scipy.optimize import linear_sum_assignment as linear_assignment
    ind = np.column_stack(linear_assignment(w.max() - w))
    return sum([w[i, j] for i, j in ind]) * 1.0 / y_pred.size

def labeled_clustering_evaluate(adata, dataset, cluster=10):

    kmeans = KMeans(n_clusters=cluster, n_init=5, random_state=200)

    adata_sol = dataset.test_sol
    # adata.obs['batch'] = adata_sol.obs['batch'][adata.obs_names]
    # adata.obs['cell_type'] = adata_sol.obs['cell_type'][adata.obs_names]
    true_labels = adata_sol.obs['cell_type'].to_numpy()

    pred_labels = kmeans.fit_predict(adata.X)
    NMI_score = round(normalized_mutual_info_score(true_labels, pred_labels, average_method='max'), 3)
    ARI_score = round(adjusted_rand_score(true_labels, pred_labels), 3)

    print('NMI: ' + str(NMI_score) + ' ARI: ' + str(ARI_score))
    return NMI_score, ARI_score


def joint_embedding_competition_score(adata, adata_sol):
    sc._settings.ScanpyConfig.n_jobs = 4
    adata.obs['batch'] = adata_sol.obs['batch'][adata.obs_names]
    adata.obs['cell_type'] = adata_sol.obs['cell_type'][adata.obs_names]
    print(adata.shape,adata_sol.shape)
    adata_bc = adata.obs_names
    adata_sol_bc = adata_sol.obs_names
    select = [item in adata_bc for item in adata_sol_bc]
    adata_sol = adata_sol[select, :]
    print(adata.shape, adata_sol.shape)

    adata.obsm['X_emb'] = adata.X
    nmi = get_nmi(adata)
    cell_type_asw = get_cell_type_ASW(adata)
    cc_con = get_cell_cycle_conservation(adata, adata_sol)
    traj_con = get_traj_conservation(adata, adata_sol)
    batch_asw = get_batch_ASW(adata)
    graph_score = get_graph_connectivity(adata)

    print('nmi %.4f, celltype_asw %.4f, cc_con %.4f, traj_con %.4f, batch_asw %.4f, graph_score %.4f\n' % (
    nmi, cell_type_asw, cc_con, traj_con, batch_asw, graph_score))

    print('average metric: %.4f' % np.mean(
        [round(i, 4) for i in [nmi, cell_type_asw, cc_con, traj_con, batch_asw, graph_score]]))

    return [round(i, 4) for i in [nmi, cell_type_asw, cc_con, traj_con, batch_asw, graph_score]]



from dance.utils import set_seed
from dance.datasets.multimodality import JointEmbeddingNIPSDataset
from dance.modules.multi_modality.joint_embedding.jae import JAEWrapper
import anndata as ad
import argparse
import random
import torch
import scanpy as sc
import numpy as np
import dgl
from dance.transforms.graph_construct import construct_basic_feature_graph, basic_feature_graph_propagation
import dance.utils.metrics as metrics

if __name__ == '__main__':
    rndseed = random.randint(0, 2147483647)
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--subtask', default = 'openproblems_bmmc_cite_phase2', choices = ['openproblems_bmmc_cite_phase2', 'openproblems_bmmc_multiome_phase2'])
    parser.add_argument('-d', '--data_folder', default = './data/joint_embedding')
    parser.add_argument('-pre', '--pretrained_folder', default = './data/joint_embedding/pretrained')
    parser.add_argument('-csv', '--csv_path', default = 'decoupled_lsi.csv')
    parser.add_argument('-seed', '--rnd_seed', default = rndseed, type=int)
    parser.add_argument('-cpu', '--cpus', default = 1, type = int)
    parser.add_argument('-device', '--device', default = 'cuda:2')
    parser.add_argument('-bs', '--batch_size', default = 128, type=int)
    parser.add_argument('-nm', '--normalize', default = 1, type=int, choices = [0, 1])

    args = parser.parse_args()

    device = args.device
    pre_normalize = bool(args.normalize)
    torch.set_num_threads(args.cpus)
    rndseed = args.rnd_seed
    set_seed(rndseed)

    dataset = JointEmbeddingNIPSDataset(args.subtask, data_dir=args.data_folder).load_data()\
        .load_metadata().load_sol().preprocess('aux', args.pretrained_folder).normalize()
    X_train, Y_train, X_test = dataset.preprocessed_data['X_train'], dataset.preprocessed_data['Y_train'], \
                                       dataset.preprocessed_data['X_test']

    model = JAEWrapper(args, dataset)
    model.fit(X_train, Y_train)
    model.load(f'models/model_joint_embedding_{rndseed}.pth')

    with torch.no_grad():
        X_test = torch.from_numpy(np.concatenate([X_train, X_test])).float().to(device)
        test_id = np.arange(X_test.shape[0])
        labels = dataset.test_sol.obs['cell_type'].to_numpy()
        embeds = model.predict(X_test, test_id).cpu().numpy()
        print(embeds)
        print(model.score(X_test, test_id, labels, 'clustering'))

    # mod1_obs = dataset.modalities[0].obs
    # mod1_uns = dataset.modalities[0].uns
    # adata = ad.AnnData(
    #     X=embeds,
    #     obs=mod1_obs,
    #     uns={
    #         'dataset_id': mod1_uns['dataset_id'],
    #         'method_id': 'scmogcn',
    #     },
    # )
    # adata.write_h5ad(f'./joint_embedding_{rndseed}.h5ad', compression="gzip")




# Copyright 2022 DSE lab.  All rights reserved.
import os
import zipfile
import tqdm
import urllib.request


# delete zip file
def delete_file(filename):
    if not os.path.exists(filename):
        print("File does not exist")
    else:
        print("Deleting", filename)
        os.remove(filename)


# download zip file from url with progress bar
def download_file(url, filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    if not os.path.exists(filename):
        u = urllib.request.urlopen(url)
        f = open(filename, "wb")
        meta = u.info()
        file_size = int(meta.get("Content-Length"))
        print(f"Downloading: {filename} Bytes: {file_size:,}")

        file_size_dl = 0
        block_sz = 8192
        with tqdm.tqdm(total=file_size, unit="B", unit_scale=True, unit_divisor=1024) as bar:
            while True:
                buffer = u.read(block_sz)
                if not buffer:
                    break
                file_size_dl += len(buffer)
                f.write(buffer)
                bar.update(len(buffer))
        f.close()
        u.close()
        return True
    else:
        print("File already downloaded")
        return False


# unzip zipfile
def unzip_file(filename, directory_to_extract_to):
    if not os.path.exists(filename):
        print("File does not exist")
    else:
        print("Unzipping", filename)
        with zipfile.ZipFile(filename, "r") as zip_ref:
            zip_ref.extractall(directory_to_extract_to)
        delete_file(filename)


def download_unzip(url, filepath):
    zip_filepath = f"{filepath}.zip"
    download_file(url, zip_filepath)
    unzip_file(zip_filepath, filepath)



import os

import pandas as pd
from sklearn.svm import SVC


class SVM():
    """The SVM cell-type classification model.

    Parameters
    ----------
    args : argparse.Namespace
        A Namespace contains arguments of SVM. See parser help document for more info.
    prj_path: str
        project path

    """

    def __init__(self, args, prj_path="./"):
        self.args = args

    def fit(self, train_labels, train_cell_feat):
        """Train the classifier.

        Parameters
        ----------
        train_labels: np.array
            Training labels.
        train_cell_feat: np.array
            Training cell features.

        """

        self.train_labels, self.train_cell_feat = train_labels, train_cell_feat
        self.model = SVC(random_state=self.args.random_seed, probability=True)
        self.model.fit(self.train_cell_feat, self.train_labels)

    def predict(self, map_dict, id2label, test_label_dict, test_feat_dict, test_cell_id_dict):
        """Predict cell labels.

        Parameters
        ----------
        map_dict: dict
            The map dictionary.
        id2label: np.array
            The dictionary for converting ID to label.
        test_label_dict: dict
            The dictionary for labels in testing set.
        test_feat_dict: dict
            The dictionary for features in testing set.
        test_cell_id_dict: dict
            The dictionary for cell ids.

        Returns
        -------
        output: dict
            A diction of predicted celllabels.

        """
        self.map_dict = map_dict
        self.id2label = id2label
        self.test_label_dict = test_label_dict
        self.test_feat_dict = test_feat_dict
        self.test_cell_id_dict = test_cell_id_dict
        output = {}
        for num in self.args.test_dataset:
            score = self.model.predict_proba(self.test_feat_dict[num])  # [cell, class-num]
            pred_labels = []
            unsure_num = correct = 0
            for pred, t_label in zip(score, self.test_label_dict[num]):
                pred_label = self.id2label[pred.argmax().item()]
                if pred_label in self.map_dict[num][t_label]:
                    correct += 1
                pred_labels.append(pred_label)
            output[num] = pred_labels
        return output

    def score(self, map_dict, id2label, test_label_dict, test_feat_dict, test_cell_id_dict):
        """Model performance score measured by accuracy.

        Parameters
        ----------
        map_dict: dict
            The map dictionary.
        id2label: dict
            The dictionary for converting ID to label.
        test_label_dict: dict
            The dictionary for labels in testing set.
        test_feat_dict: dict
            The dictionary for features in testing set.
        test_cell_id_dict: dict
            The dictionary for cell ids.

        Returns
        -------
        accuracy_all: dict
            A diction of accuracy on different testing sets.

        """
        self.map_dict = map_dict
        self.id2label = id2label
        self.test_label_dict = test_label_dict
        self.test_feat_dict = test_feat_dict
        self.test_cell_id_dict = test_cell_id_dict
        accuracy_all = {}
        for num in self.args.test_dataset:
            score = self.model.predict_proba(self.test_feat_dict[num])  # [cell, class-num]
            pred_labels = []
            unsure_num = correct = 0
            for pred, t_label in zip(score, self.test_label_dict[num]):
                pred_label = self.id2label[pred.argmax().item()]
                if pred_label in self.map_dict[num][t_label]:
                    correct += 1
                pred_labels.append(pred_label)

            acc = correct / score.shape[0]
            print(f"SVM-{self.args.species}-{self.args.tissue}-{num}-ACC: {acc:.5f}")
            accuracy_all[num] = acc
        return accuracy_all

    def save(self, num, pred):
        """Save the predictions.

        Parameters
        ----------
        num: int
            test file name
        pred: dict
            prediction labels

        """
        label_map = pd.read_excel(self.prj_path / "data" / "celltype2subtype.xlsx", sheet_name=self.args.species,
                                  header=0, names=["species", "old_type", "new_type", "new_subtype"])

        save_path = self.prj_path / self.args.save_dir
        if not save_path.exists():
            save_path.mkdir()

        label_map = label_map.fillna("N/A", inplace=False)
        oldtype2newtype = {}
        oldtype2newsubtype = {}
        for _, old_type, new_type, new_subtype in label_map.itertuples(index=False):
            oldtype2newtype[old_type] = new_type
            oldtype2newsubtype[old_type] = new_subtype
        if not os.path.exists(self.args.save_dir):
            os.mkdir(self.args.save_dir)

        df = pd.DataFrame({
            "index": self.test_cell_id_dict[num],
            "original label": self.test_label_dict[num],
            "cell type": [oldtype2newtype.get(p, p) for p in pred],
            "cell subtype": [oldtype2newsubtype.get(p, p) for p in pred]
        })
        df.to_csv(save_path / ("SVM_" + self.args.species + f"_{self.args.tissue}_{num}.csv"), index=False)
        print(f"output has been stored in {self.args.species}_{self.args.tissue}_{num}.csv")



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributions as dist
from torch import optim
import math
import numpy as np
from numpy import prod, sqrt
from pyro.distributions.zero_inflated import ZeroInflatedNegativeBinomial
from torch.utils.data import DataLoader
from sklearn.cluster import KMeans, DBSCAN
from dance.utils import SimpleIndexDataset
import os
import scipy.sparse as sp
import pandas as pd

def get_mean(d, K=100):
    """
    Extract the `mean` parameter for given distribution.
    If attribute not available, estimate from samples.
    """
    try:
        mean = d.mean
    except NotImplementedError:
        samples = d.rsample(torch.Size([K]))
        mean = samples.mean(0)
    return mean

def kl_divergence(d1, d2, K=100):
    """Computes closed-form KL if available, else computes a MC estimate."""
    if (type(d1), type(d2)) in torch.distributions.kl._KL_REGISTRY:
        return torch.distributions.kl_divergence(d1, d2)
    else:
        samples = d1.rsample(torch.Size([K]))
        return (d1.log_prob(samples) - d2.log_prob(samples)).mean(0)

def m_elbo_naive(model, x):
    """Computes E_{p(x)}[ELBO] for multi-modal vae --- NOT EXPOSED"""
    qz_xs, px_zs, zss = model(x)
    lpx_zs, klds = [], []
    for r, qz_x in enumerate(qz_xs):
        kld = kl_divergence(qz_x, model.pz(*model._get_pz_params))
        klds.append(kld.sum(-1))
        for d, px_z in enumerate(px_zs[r]):
            lpx_z = px_z.log_prob(x[d]) * model.vaes[d].llik_scaling
            lpx_zs.append(lpx_z.sum(-1))
    obj = (1 / len(model.vaes)) * (torch.stack(lpx_zs).sum(0) - torch.stack(klds).sum(0))
    return obj.sum()

def m_elbo_naive_warmup(model, x, beta):
    """Computes E_{p(x)}[ELBO] for multi-modal vae --- NOT EXPOSED"""
    qz_xs, px_zs, zss = model(x)
    lpx_zs, klds = [], []
    for r, qz_x in enumerate(qz_xs):
        kld = kl_divergence(qz_x, model.pz(*model._get_pz_params))
        klds.append(kld.sum(-1))
        for d, px_z in enumerate(px_zs[r]):
            lpx_z = px_z.log_prob(x[d]) * model.vaes[d].llik_scaling
            lpx_zs.append(lpx_z.sum(-1))
    obj = (1 / len(model.vaes)) * (torch.stack(lpx_zs).sum(0) - beta * torch.stack(klds).sum(0))
    return obj.sum()

def protein_preprocessing(t1):
    t0 = t1.clone()
    t0[t0 == 0] = 1
    return torch.log1p(t1 / torch.exp(torch.sum(torch.log(t0), axis=1) * (1 / torch.sum(t1 > 0, axis=1))).unsqueeze(-1))

def atac_preprocessing(t1):
    t1[t1>0] = 1
    return t1

# TODO: Not implemented
def rna_preprocessing(t1):
    return t1

class Constants(object):
    eta = 1e-6
    eps = 1e-7
    log2 = math.log(2)
    log2pi = math.log(2 * math.pi)
    logceilc = 88  # largest cuda v s.t. exp(v) < inf
    logfloorc = -104  # smallest cuda v s.t. exp(v) > 0

class ZINB(ZeroInflatedNegativeBinomial):
    def __init__(self, total_count, probs, gate):
        super(ZINB, self).__init__(total_count=total_count, probs=probs, gate=gate)

class VAE(nn.Module):
    def __init__(self, prior_dist, likelihood_dist, post_dist, enc, dec, params):
        super(VAE, self).__init__()
        self.pz = prior_dist
        self.px_z = likelihood_dist
        self.qz_x = post_dist
        self.enc = enc
        self.dec = dec
        self.modelName = None
        self.params = params
        self._pz_params = None  # defined in subclass
        self._qz_x_params = None  # populated in `forward`
        self.llik_scaling = 1.0

    @property
    def pz_params(self):
        return self._pz_params

    @property
    def qz_x_params(self):
        if self._qz_x_params is None:
            raise NameError("qz_x params not initalised yet!")
        return self._qz_x_params

    @staticmethod
    def getDataLoaders(batch_size, shuffle=True, device="cuda"):
        # handle merging individual datasets appropriately in sub-class
        raise NotImplementedError

    def forward(self, x):
        self._qz_x_params = self.enc(x)
        qz_x = self.qz_x(*self._qz_x_params)
        zs = qz_x.rsample()
        px_z = self.px_z(*self.dec(zs))
        return qz_x, px_z, zs

    def reconstruct(self, data):
        self.eval()
        with torch.no_grad():
            qz_x = self.qz_x(*self.enc(data))
            latents = qz_x.rsample()  # no dim expansion
            px_z = self.px_z(*self.dec(latents))
            recon = get_mean(px_z)
        return recon

    def reconstruct_sample(self, data):
        self.eval()
        with torch.no_grad():
            qz_x = self.qz_x(*self.enc(data))
            latents = qz_x.rsample()  # no dim expansion
            px_z = self.px_z(*self.dec(latents))
            recon = px_z._sample()
        return recon

    def latents(self, data, sampling=False):
        self.eval()
        with torch.no_grad():
            qz_x = self.qz_x(*self.enc(data))
            if not sampling:
                lats = get_mean(qz_x)
            else:
                lats = qz_x._sample()
        return lats

class Enc(nn.Module):

    def __init__(self, data_dim, latent_dim, num_hidden_layers, hidden_dim):  # added hidden_dim
        super(Enc, self).__init__()
        self.data_dim = data_dim
        modules = []
        modules.append(nn.Sequential(nn.Linear(data_dim, hidden_dim), nn.BatchNorm1d(hidden_dim), nn.ReLU(True)))
        for _ in range(num_hidden_layers - 1):
            modules.append(nn.Sequential(nn.Linear(hidden_dim, hidden_dim), nn.BatchNorm1d(hidden_dim), nn.ReLU(True)))
        self.enc = nn.Sequential(*modules)
        self.fc21 = nn.Linear(hidden_dim, latent_dim)
        self.fc22 = nn.Linear(hidden_dim, latent_dim)
        self.scale_factor = 10000

    def read_count(self, x):
        read = torch.sum(x, axis=1)
        read = read.repeat(self.data_dim, 1).t()
        return (read)

    def forward(self, x):
        read = self.read_count(x)
        x = x / read * self.scale_factor
        e = self.enc(x)
        lv = self.fc22(e).clamp(-12, 12)  # restrict to avoid torch.exp() over/underflow
        return self.fc21(e), F.softmax(lv, dim=-1) * lv.size(-1) + Constants.eta

class Dec(nn.Module):
    """ Generate an MNIST image given a sample from the latent space. """

    def __init__(self, data_dim, latent_dim, num_hidden_layers, hidden_dim, modality):  # added hidden_dim
        super(Dec, self).__init__()
        self.modality = modality
        self.data_dim = data_dim

        modules = []
        modules.append(nn.Sequential(nn.Linear(latent_dim, hidden_dim), nn.BatchNorm1d(hidden_dim), nn.ReLU(True)))
        for _ in range(num_hidden_layers - 1):
            modules.append(nn.Sequential(nn.Linear(hidden_dim, hidden_dim), nn.BatchNorm1d(hidden_dim), nn.ReLU(True)))
        self.dec = nn.Sequential(*modules)
        self.fc31 = nn.Linear(hidden_dim, data_dim)
        self.fc32 = nn.Linear(hidden_dim, data_dim)

        if self.modality == 'atac':
            # zero-inflated
            self.fc33 = nn.Linear(hidden_dim, data_dim)

    def forward(self, z):
        d = self.dec(z)
        log_r = self.fc31(d).clamp(-12, 12)  # restrict to avoid torch.exp() over/underflow
        r = torch.exp(log_r)
        p = self.fc32(d)
        p = torch.sigmoid(p).clamp(Constants.eps, 1 - Constants.eps)  # restrict to avoid probs = 0,1

        if self.modality == 'atac':
            g = self.fc33(d)
            g = torch.sigmoid(g)
            return r, p, g
        else:
            return r, p

class ATAC(VAE):
    """ Derive a specific sub-class of a VAE for ATAC. """

    def __init__(self, params):
        super(ATAC, self).__init__(
            dist.Laplace,
            ZINB,
            dist.Laplace,
            Enc(params.p_dim, params.latent_dim, params.num_hidden_layers, params.p_hidden_dim),
            Dec(params.p_dim, params.latent_dim, params.num_hidden_layers, params.p_hidden_dim, 'atac'),
            params
        )
        grad = {'requires_grad': params.learn_prior}
        self._pz_params = nn.ParameterList([
            nn.Parameter(torch.zeros(1, params.latent_dim), requires_grad=False),  # mu
            nn.Parameter(torch.zeros(1, params.latent_dim), **grad)  # logvar
        ])
        self.modelName = 'atac'
        self.data_dim = self.params.p_dim
        self.llik_scaling = 1.
        self.scale_factor = 10000

    @property
    def pz_params(self):
        return self._pz_params[0], F.softmax(self._pz_params[1], dim=1) * self._pz_params[1].size(-1)

    @staticmethod
    def getDataLoaders(dataset, batch_size, shuffle=True, device="cuda"):
        kwargs = {'num_workers': 1, 'pin_memory': True} if device == "cuda" else {}
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, drop_last=False, **kwargs)
        return dataloader

    def forward(self, x):
        read_count = self.enc.read_count(x)
        self._qz_x_params = self.enc(x)
        qz_x = self.qz_x(*self._qz_x_params)
        zs = qz_x.rsample()
        r, p, g = self.dec(zs)
        r = r / self.scale_factor * read_count
        px_z = self.px_z(r, p, g)
        return qz_x, px_z, zs

class Protein(VAE):
    """ Derive a specific sub-class of a VAE for Protein. """
    def __init__(self, params):
        super(Protein, self).__init__(
            dist.Laplace,
            dist.NegativeBinomial,
            dist.Laplace,
            Enc(params.p_dim, params.latent_dim, params.num_hidden_layers, params.p_hidden_dim),
            Dec(params.p_dim, params.latent_dim, params.num_hidden_layers, params.p_hidden_dim, 'protein'),
            params
        )
        grad = {'requires_grad': params.learn_prior}
        self._pz_params = nn.ParameterList([
            nn.Parameter(torch.zeros(1, params.latent_dim), requires_grad=False),  # mu
            nn.Parameter(torch.zeros(1, params.latent_dim), **grad)  # logvar
        ])
        self.modelName = 'protein'
        self.data_dim = self.params.p_dim
        self.llik_scaling = 1.
        self.scale_factor = 10000

    @property
    def pz_params(self):
        return self._pz_params[0], F.softmax(self._pz_params[1], dim=1) * self._pz_params[1].size(-1)

    @staticmethod
    def getDataLoaders(dataset, batch_size, shuffle=True, device="cuda"):
        kwargs = {'num_workers': 1, 'pin_memory': True} if device == "cuda" else {}
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, drop_last=False, **kwargs)

        return dataloader

    def forward(self, x):
        read_count = self.enc.read_count(x)
        self._qz_x_params = self.enc(x)
        qz_x = self.qz_x(*self._qz_x_params)
        zs = qz_x.rsample()
        r, _ = self.dec(zs)
        r = r / self.scale_factor * read_count
        px_z = self.px_z(r, _)
        return qz_x, px_z, zs

class RNA(VAE):
    """ Derive a specific sub-class of a VAE for RNA. """

    def __init__(self, params):
        super(RNA, self).__init__(
            dist.Laplace,
            dist.NegativeBinomial,  # likelihood
            dist.Laplace,
            Enc(params.r_dim, params.latent_dim, params.num_hidden_layers, params.r_hidden_dim),
            Dec(params.r_dim, params.latent_dim, params.num_hidden_layers, params.r_hidden_dim, 'rna'),
            params
        )
        grad = {'requires_grad': params.learn_prior}
        self._pz_params = nn.ParameterList([
            nn.Parameter(torch.zeros(1, params.latent_dim), requires_grad=False),  # mu
            nn.Parameter(torch.zeros(1, params.latent_dim), **grad)  # logvar
        ])
        self.modelName = 'rna'
        self.data_dim = self.params.r_dim
        self.llik_scaling = 1.
        self.scale_factor = 10000

    @property
    def pz_params(self):
        return self._pz_params[0], F.softmax(self._pz_params[1], dim=1) * self._pz_params[1].size(-1)

    @staticmethod
    def getDataLoaders(dataset, batch_size, shuffle=True, device="cuda"):
        kwargs = {'num_workers': 1, 'pin_memory': True} if device == "cuda" else {}
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, drop_last=False, **kwargs)

        return dataloader

    def forward(self, x):
        read_count = self.enc.read_count(x)
        self._qz_x_params = self.enc(x)
        qz_x = self.qz_x(*self._qz_x_params)
        zs = qz_x.rsample()
        r, _ = self.dec(zs)
        r = r / self.scale_factor * read_count
        px_z = self.px_z(r, _)
        return qz_x, px_z, zs

class MMVAE(nn.Module):
    """MMVAE class.

    Parameters
    ----------
    subtask : str
        Name of the subtask which is composed of the name of two modality. This parameter will indicate some modality-specific features in the model.
    params : argparse.Namespace
        A Namespace object that contains arguments of MMVAE. For details of parameters in parser args, please refer to link (parser help document).
    """

    def __init__(self, subtask, params):
        super(MMVAE, self).__init__()
        self.pz = dist.Laplace
        assert subtask in ('rna-dna', 'rna-protein')
        self.modelName = subtask
        if subtask == 'rna-dna':
            self.preprocessing = atac_preprocessing
            self.vaes = nn.ModuleList([RNA(params), ATAC(params)])
        else:
            self.preprocessing = protein_preprocessing
            self.vaes = nn.ModuleList([RNA(params), Protein(params)])
        self.params = params

        grad = {'requires_grad': params.learn_prior}
        self._pz_params = nn.ParameterList([
            nn.Parameter(torch.zeros(1, params.latent_dim), requires_grad=False),  # mu
            nn.Parameter(torch.zeros(1, params.latent_dim), **grad)  # logvar
        ])
        self.vaes[0].llik_scaling = prod(self.vaes[1].dataSize) / prod(self.vaes[0].dataSize) \
            if params.llik_scaling == 0 else params.llik_scaling
        self.scale_factor = 10000


    @property
    def _get_pz_params(self):
        return self._pz_params[0], F.softmax(self._pz_params[1], dim=1) * self._pz_params[1].size(-1)

    def _get_cluster(self, data, modality='both', n_clusters=10, method='kmeans', device='cuda'):
        self.eval()
        lats = self._get_latents(data, sampling=False)
        if modality == 'both':
            lat = sum(lats) / len(lats)
        elif modality == 'rna':
            lat = lats[0]
        elif modality == 'atac':
            lat = lats[1]

        if method == 'kmeans':
            fit = KMeans(n_clusters=n_clusters, random_state=0, init='k-means++').fit(lat.cpu().numpy())
            cluster = fit.labels_
        elif method == 'dbscan':
            fit = DBSCAN(eps=0.5, min_samples=50).fit(lat.cpu().numpy())
            cluster = fit.labels_
        else:
            gamma, _, _, _, _ = self.get_gamma(lat)
            cluster = torch.argmax(gamma, axis=1)
            cluster = cluster.detach().numpy()
            fit = None

        return cluster, fit

    def forward(self, x):
        """Forward function for torch.nn.Module.

        Parameters
        ----------
        x : list[torch.Tensor]
            Features of two modalities.

        Returns
        -------
        qz_xs : list[torch.Tensor]
            Post prior of two modalities.
        px_zs : list[torch.Tensor]
            likelihood of two modalities.
        zss : list[torch.Tensor]
            Reconstruction results of two modalities.
        """
        qz_xs, zss = [], []
        read_counts = []
        # initialise cross-modal matrix
        px_zs = [[None for _ in range(len(self.vaes))] for _ in range(len(self.vaes))]
        for m, vae in enumerate(self.vaes):
            read_counts.append(vae.enc.read_count(x[m]))
            qz_x, px_z, zs = vae(x[m])
            qz_xs.append(qz_x)
            zss.append(zs)
            px_zs[m][m] = px_z  # fill-in diagonal
        for e, zs in enumerate(zss):
            for d, vae in enumerate(self.vaes):
                if e != d:  # fill-in off-diagonal
                    if self.modelName == 'rna-protein':
                        r, _ = vae.dec(zs)
                        r = r / self.scale_factor * read_counts[d]
                        px_zs[e][d] = vae.px_z(r, _)
                    else:
                        if d == 0:
                            r, p = vae.dec(zs)
                            r = r / self.scale_factor * read_counts[d]
                            px_zs[e][d] = vae.px_z(r, p)
                        else:
                            r, p, g = vae.dec(zs)
                            r = r / self.scale_factor * read_counts[d]
                            px_zs[e][d] = vae.px_z(r, p, g)

        return qz_xs, px_zs, zss

    def _reconstruct(self, data):
        self.eval()
        with torch.no_grad():
            _, px_zs, _ = self.forward(data)
            # cross-modal matrix of reconstructions of MEANS
            recons = [[get_mean(px_z) for px_z in r] for r in px_zs]
        return recons

    def _reconstruct_sample(self, data):
        self.eval()
        with torch.no_grad():
            _, px_zs, _ = self.forward(data)
            # cross-modal matrix of reconstructions of SAMPLES
            recons = [[px_z.sample() for px_z in r] for r in px_zs]
        return recons

    def _get_latents(self, data, sampling=False):
        self.eval()
        with torch.no_grad():
            qz_xs, _, _ = self.forward(data)
            if not sampling:
                lats = [get_mean(qz_x) for qz_x in qz_xs]
            else:
                lats = [qz_x._sample() for qz_x in qz_xs]
        return lats

    def fit(self, dataset):
        """fit function for training.

        Parameters
        ----------
        dataset : dance.datasets.multimodality.ModalityPredictionDataset
            Dataset for modality prediction.

        Returns
        -------
        None.
        """

        start_early_stop = self.params.deterministic_warmup

        idx = np.random.permutation(dataset.modalities[0].shape[0])
        train_idx = idx[:int(idx.shape[0] * 0.85)]
        val_idx = idx[int(idx.shape[0] * 0.85):]

        optimizer = optim.Adam(filter(lambda p: p.requires_grad, self.parameters()),
                               lr=self.params.lr, amsgrad=True)
        assert (self.params.obj in ['m_elbo_naive', 'm_elbo_naive_warmup'])
        objective = m_elbo_naive_warmup if self.params.obj == 'm_elbo_naive_warmup' else 'm_elbo_naive'
        train_dataset = SimpleIndexDataset(train_idx)
        train_loader = DataLoader(
            dataset=train_dataset,
            batch_size=self.params.batch_size,
            shuffle=True,
            num_workers=0,
            drop_last=True,
        )

        train_mod1 = torch.from_numpy(dataset.numpy_features(0, True)).float().to(self.params.device)
        train_mod2 = torch.from_numpy(dataset.numpy_features(1, True)).float().to(self.params.device)
        self.ratio = train_mod2.sum()/train_mod1.sum()
        vals = []
        tr = []

        for epoch in range(1, self.params.epochs + 1):
            self.train()
            b_loss = 0
            for i, batch_idx in enumerate(train_loader):
                dataT = (train_mod1[batch_idx], train_mod2[batch_idx])
                beta = (epoch - 1) / start_early_stop if epoch <= start_early_stop else 1
                if dataT[0].size()[0] == 1:
                    continue
                # data = [d.to(self.paradevice) for d in dataT]  # multimodal
                data = dataT
                optimizer.zero_grad()
                loss = -objective(self, data, beta)
                loss.backward()
                optimizer.step()
                b_loss += loss.item()
                if self.params.print_freq > 0 and i % self.params.print_freq == 0:
                    print("iteration {:04d}: loss: {:6.3f}".format(i, loss.item() / self.params.batch_size))
            tr.append(b_loss / len(train_loader.dataset))
            print('====> Epoch: {:03d} Train loss: {:.4f}'.format(epoch, tr[-1]))

            if torch.isnan(torch.tensor([b_loss])):
                break

            vals.append(self.score(train_mod1[val_idx], train_mod2[val_idx], metric='loss'))
            print('====>             Valid loss: {:.4f}'.format(vals[-1]))

            if vals[-1] == min(vals):
                if not os.path.exists('models'):
                    os.mkdir('models')
                torch.save(self.state_dict(), f'models/model_{self.params.rnd_seed}.pth')

            if epoch % 10 == 0:
                print('Valid RMSELoss:', self.score(train_mod1[val_idx], train_mod2[val_idx]))

            if epoch > start_early_stop and min(vals) != min(vals[-10:]):
                print('Early stopped.')
                self.load_state_dict(torch.load(f'models/model_{self.params.rnd_seed}.pth'))
                break

    def score(self, X, Y, metric ='rmse'):
        """Score function to get score of prediction.

        Parameters
        ----------
        X : torch.Tensor
            Features of input modality.
        Y : torch.Tensor
            Features of input modality.
        metric : str optional
            Metric of the score function, by default to be 'rmse'.

        Returns
        -------
        score : float
            Score of predicted matching, according to specified metric.

        """
        self.eval()

        if metric == 'loss':
            b_loss = 0
            idx = np.arange(X.shape[0])
            dataset = SimpleIndexDataset(idx)
            data_loader = DataLoader(
                dataset=dataset,
                batch_size=self.params.batch_size,
                shuffle=False,
                num_workers=0,
                drop_last=False,
            )
            with torch.no_grad():
                for i, batch_idx in enumerate(data_loader):
                    objective = m_elbo_naive_warmup if self.params.obj == 'm_elbo_naive_warmup' else 'm_elbo_naive'
                    loss = -objective(self, [X[batch_idx], Y[batch_idx]], 1).item()
                    b_loss += loss
            return b_loss / X.shape[0]
        elif metric == 'rmse':
            mse = nn.MSELoss()
            pred_test = self.predict(X)
            pred = self.preprocessing(pred_test)
            pred = torch.nan_to_num(pred)
            label = self.preprocessing(Y)
            return math.sqrt(mse(pred, label).item())
        else:
            print('Warning: undefined evaluation metric.')

    #TODO: predict function need to be modified to remove label input
    def predict(self, X):
        """Score function to get score of prediction.

        Parameters
        ----------
        X : torch.Tensor
            Features of input modality and target modality.

        Returns
        -------
        pred : torch.Tensor
            Prediction of target modality from input modality.

        """
        self.eval()
        idx = np.arange(X.shape[0])
        dataset = SimpleIndexDataset(idx)
        data_loader = DataLoader(
            dataset=dataset,
            batch_size=self.params.batch_size*10,
            shuffle=False,
            num_workers=0,
            drop_last=False,
        )
        # uni, cross = [], []
        pred = []
        with torch.no_grad():
            for i, batch_idx in enumerate(data_loader):
                read_count = self.vaes[1].enc.read_count(X[batch_idx] * self.ratio)
                qz_x, px_z, zs = self.vaes[0](X[batch_idx])
                if self.modelName == 'rna-protein':
                    r, _ = self.vaes[1].dec(zs)
                    r = r / self.scale_factor * read_count
                    pred.append(self.vaes[1].px_z(r, _).sample())
                else:
                    r, p, g = self.vaes[1].dec(zs)
                    r = r / self.scale_factor * read_count
                    pred.append(self.vaes[1].px_z(r, p, g).sample())
        pred = torch.cat(pred, 0)
        return pred

        #         recons_mat = self._reconstruct_sample(dataT)
        #         for e, recons_list in enumerate(recons_mat):
        #             for d, recon in enumerate(recons_list):
        #                 if e == d:
        #                     recon = recon.cpu()
        #                     if i == 0:
        #                         uni.append(recon)
        #                     else:
        #                         uni[e] = torch.cat([uni[e], recon])
        #                 else:
        #                     recon = recon.cpu()
        #                     if i == 0:
        #                         cross.append(recon)
        #                     else:
        #                         cross[e] = torch.cat((cross[e], recon))
        # return uni, cross



from dance.utils import set_seed
from dance.datasets.multimodality import JointEmbeddingNIPSDataset
from dance.modules.multi_modality.joint_embedding.scmogcn import ScMoGCNWrapper
import anndata as ad
import argparse
import random
import torch
import scanpy as sc
import numpy as np
import dgl
from dance.transforms.graph_construct import construct_basic_feature_graph, basic_feature_graph_propagation
import dance.utils.metrics as metrics

if __name__ == '__main__':
    rndseed = random.randint(0, 2147483647)
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--subtask', default = 'openproblems_bmmc_cite_phase2', choices = ['openproblems_bmmc_cite_phase2', 'openproblems_bmmc_multiome_phase2'])
    parser.add_argument('-d', '--data_folder', default = './data/joint_embedding')
    parser.add_argument('-pre', '--pretrained_folder', default = './data/joint_embedding/pretrained')
    parser.add_argument('-csv', '--csv_path', default = 'decoupled_lsi.csv')
    parser.add_argument('-l', '--layers', default = 3, type=int, choices = [3,4,5,6,7])
    parser.add_argument('-dis', '--disable_propagation', default = 0, type=int, choices = [0, 1, 2])
    parser.add_argument('-seed', '--rnd_seed', default = rndseed, type=int)
    parser.add_argument('-cpu', '--cpus', default = 1, type = int)
    parser.add_argument('-device', '--device', default = 'cuda:2')
    parser.add_argument('-bs', '--batch_size', default = 512, type=int)
    parser.add_argument('-nm', '--normalize', default = 1, type=int, choices = [0, 1])

    args = parser.parse_args()

    device = args.device
    pre_normalize = bool(args.normalize)
    torch.set_num_threads(args.cpus)
    rndseed = args.rnd_seed
    set_seed(rndseed)

    dataset = JointEmbeddingNIPSDataset(args.subtask, data_dir=args.data_folder).load_data()\
        .load_metadata().load_sol().preprocess('aux', args.pretrained_folder).normalize()
    X_train, Y_train, X_test = dataset.preprocessed_data['X_train'], dataset.preprocessed_data['Y_train'], \
                                       dataset.preprocessed_data['X_test']

    g = construct_basic_feature_graph(X_train, X_test, device=device)
    X = basic_feature_graph_propagation(g, layers=args.layers, device=device)

    l = args.layers - 1

    model = ScMoGCNWrapper(args, dataset)
    model.fit(dataset, X, Y_train)
    model.load(f'models/model_joint_embedding_{rndseed}.pth')

    with torch.no_grad():
        test_id = np.arange(X[0].shape[0])
        labels = dataset.test_sol.obs['cell_type'].to_numpy()
        embeds = model.predict(X, test_id).cpu().numpy()
        print(embeds)
        print(model.score(X, test_id, labels, 'clustering'))

    # mod1_obs = dataset.modalities[0].obs
    # mod1_uns = dataset.modalities[0].uns
    # adata = ad.AnnData(
    #     X=embeds,
    #     obs=mod1_obs,
    #     uns={
    #         'dataset_id': mod1_uns['dataset_id'],
    #         'method_id': 'scmogcn',
    #     },
    # )
    # adata.write_h5ad(f'./joint_embedding_{rndseed}.h5ad', compression="gzip")
    #
    # # old Neurips metrics
    # # scmogcn_test(adata)
    # # scmvae test
    # metrics.labeled_clustering_evaluate(adata, dataset)


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import optim
from torch.autograd import Variable
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class MSLELoss(nn.Module):
    """MSLELoss.

        Parameters
        ----------
        None

        Returns
        -------
        None. 

        """
    def __init__(self):
        super().__init__()
        self.mse = nn.MSELoss()
        
    def forward(self, pred, actual):
        """forward function.

        Parameters
        ----------
        pred : torch tensor
            linear transformation of cell profile (reference basis) matrix.
        actual : torch tensor
            mixture expression matrix.

        Returns
        -------
        loss : float
            mean squared log error loss.

        """
        pred_clamp = pred.clamp(min=0)
        actual_clamp = actual.clamp(min=0)
        loss = self.mse(torch.log1p(pred_clamp), torch.log1p(actual_clamp))
        return loss
        

class SpatialDecon(object):
    """SpatialDecon.

        Parameters
        ----------
        in_dim : int 
            input dimension.
        out_dim : int 
            output dimension.
        bias : boolean optional
            include bias term, default False.

        Returns
        -------
        None. 

        """
    def __init__(self,in_dim, out_dim, bias=False, init_bias=None, device = torch.device("cuda" if torch.cuda.is_available() else "cpu")):
        super(SpatialDecon, self).__init__()
        self.device = device
        self.model = nn.Linear(in_features=in_dim, out_features=out_dim, bias=bias)
        self.model.bias = init_bias
        self.model = self.model.to(device)
        
    def forward(self, x:torch.Tensor):
        """forward function.

        Parameters
        ----------
        x : torch tensor
            input features.

        Returns
        -------
        output : torch tensor
            linear projection of input.

        """
        out = self.model(x)
        return(out)
    
    def predict(self):
        """prediction function.
        Parameters
        ----------
     
        Returns
        -------
        
        proportion_preds : torch tensor
            predictions of cell-type proportions.
        
        """
        proportion_preds = self.model.weight.T
        proportion_preds = proportion_preds / torch.sum(proportion_preds, axis=0)
        return(proportion_preds)
    
    def fit(self,ref_x, y, max_iter, lr, print_res=False, print_period=100):
        """fit function for model training.

        Parameters
        ----------
        ref_x : 
            scRNA-seq reference expression.
        y : 
            mixed cell expression.
        max_iter : int
            maximum number of iterations for optimizat.
        lr : float
            learning rate.
        print_res : bool optional 
            indicates to print live training results, default False.
        print_period : int optional
            indicates number of iterations until training results print. 
            
        Returns
        -------
        None.

        """
        
        ref_x = Variable(torch.FloatTensor(ref_x), requires_grad=True).to(self.device)
        y = Variable(torch.FloatTensor(y)).to(self.device)
        
        criterion = MSLELoss()
        optimizer=optim.Adam(self.model.parameters(), lr=lr)
        
        self.model.train()
        for iteration in range(max_iter):
            iteration += 1
            y_pred = self.model(ref_x)

            # Compute and print loss
            loss = criterion(y_pred, y)
            self.loss = loss
            # Zero gradients, perform a backward pass,
            # and update the weights.
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

                
            with torch.no_grad():
                self.model.weight.copy_(self.model.weight.data.clamp(min=0))
            if iteration % print_period  == 0:
                print(f"Epoch: {iteration:02}/{max_iter} Loss: {loss.item():.5e}")
                
    def score(self, ref_x, y):
        """score.

        Parameters
        ----------
        ref_x : 
            scRNA reference expression
        y : 
            cell-mixture expression

        Returns
        -------
        model_score : float
            MSLE loss between transformed scRNA reference expression (prediction) and cell-mixture expression.

        """
        ref_x = torch.FloatTensor(ref_x).to(self.device)
        y = torch.FloatTensor(y).to(self.device)
        y_pred = self.model(ref_x)
        
        criterion = MSLELoss()
        model_score = criterion(y_pred, y).item()
        return model_score


"""Reimplementation of the ACTINN cell-type annotation method.

Reference
---------
Ma, Feiyang, and Matteo Pellegrini. "ACTINN: automated identification of cell types in single cell RNA sequencing."
Bioinformatics 36.2 (2020): 533-538.

"""
import itertools
from typing import Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


class ACTINN(nn.Module):
    """The ACTINN cell-type classification model."""

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        *,
        batch_size: int = 128,
        device: str = "cpu",
        hidden_dims: Tuple[int, ...] = (100, 50, 25),
        lambd: float = 0.01,
        lr: float = 0.01,
        num_epochs: int = 50,
        print_cost: bool = False,
    ):
        """Initialize the ACTINN model.

        Parameters
        ----------
        input_dim : int
            Input dimension (number of genes)
        output_dim : int
            Output dimension (number of cell types)
        hidden_dims : :obj:`tuple` of int
            Hidden layer dimensions.
        batch_size : int
            Training batch size
        device : str
            Training device
        lambd : float
            Regularization parameter
        lr : float
            Initial learning rate
        num_epochs : int
            Number of epochs to run
        print_cost : bool
            Print training loss if set to True

        """
        super().__init__()

        # Save attributes
        self.batch_size = batch_size
        self.device = device
        self.lambd = lambd
        self.lr = lr
        self.num_epochs = num_epochs
        self.print_cost = print_cost

        # Build MLP
        self.model = nn.Sequential(
            nn.Linear(input_dim, hidden_dims[0]),
            nn.ReLU(),
            *itertools.chain.from_iterable(
                zip(
                    map(nn.Linear, hidden_dims[:-1], hidden_dims[1:]),
                    itertools.repeat(nn.ReLU()),
                )
            ),
            nn.Linear(hidden_dims[-1], output_dim),
        ).to(device)
        print(self.model)

    def forward(self, x):
        """Forward propagation."""
        return self.model(x)

    @torch.no_grad()
    def initialize_parameters(self, seed=None):
        """Initialize parameters."""
        if seed is not None:
            torch.manual_seed(seed)

        for i in range(0, len(self.model), 2):
            nn.init.xavier_normal_(self.model[i].weight)
            self.model[i].bias[:] = 0

    def compute_loss(self, z, y):
        """Compute loss function.

        Parameters
        ----------
        z : torch.Tensor
            Output of forward propagation (cells by cell-types).
        y : torch.Tensor
            Cell labels (cells).

        Returns
        -------
        torch.Tensor
            Loss.

        """
        log_prob = F.log_softmax(z, dim=-1)
        loss = nn.NLLLoss()(log_prob, y)
        for i in range(0, len(self.model), 2):  # TODO: replace with weight_decay
            loss += self.lambd * torch.sum(self.model[i].weight ** 2) / 2

        if self.print_cost:
            print(loss)

        return loss

    def random_batches(self, x, y, batch_size=32, seed=None):
        """Shuffle data and split into batches.

        Parameters
        ----------
        x : torch.Tensor
            Training data (cells by genes).
        y : torch.Tensor
            True labels (cells by cell-types).

        Yields
        ------
        Tuple[torch.Tensor, torch.Tensor]
            Batch of training data (x, y).

        """
        ns = x.shape[0]
        perm = np.random.default_rng(seed).permutation(ns).tolist()
        slices = [perm[i: i + batch_size] for i in range(0, ns, batch_size)]
        yield from map(lambda slice_: (x[slice_], y[slice_]), slices)

    def fit(self, x_train, y_train, seed=None):
        """Fit the classifier.

        Parameters
        ----------
        x_train : torch.Tensor
            training data (genes by cells).
        y_train : torch.Tensor
            training labels (cell-types by cells).
        seed : int, optional
            Random seed, if set to None, then random.

        """
        x_train = x_train.T.clone().detach().float().to(self.device)  # cells by genes
        y_train = torch.where(y_train.T)[1].to(self.device)  # cells

        # Initialize weights, optimizer, and scheduler
        self.initialize_parameters(seed)
        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr)
        lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=optimizer, gamma=0.95)

        # Start training loop
        for epoch in range(self.num_epochs):
            epoch_seed = seed if seed is None else seed + epoch
            batches = self.random_batches(x_train, y_train, self.batch_size, epoch_seed)

            for batch_x, batch_y in batches:
                batch_cost = self.compute_loss(self.forward(batch_x), batch_y)

                optimizer.zero_grad()
                batch_cost.backward()
                optimizer.step()
                lr_scheduler.step()

        print("Parameters have been trained!")

    @torch.no_grad()
    def predict(self, x):
        """Predict cell labels.

        Parameters
        ----------
        x : torch.Tensor
            Gene expression input features (genes by cells).

        Returns
        -------
        torch.Tensor
            Predicted cell-label indices.

        """
        x = x.T.clone().detach().to(self.device)
        z = self.forward(x)
        prediction = torch.argmax(z, dim=-1)
        return prediction

    def score(self, x, y):
        """Model performance score measured by accuracy.

        Parameters
        ----------
        x : torch.Tensor
            Gene expression input features (genes by cells).
        y : torch.Tensor
            One-hot encoded ground truth labels (cell-types by cells).

        Returns
        -------
        float
            Prediction accuracy

        """
        pred = self.predict(x).detach().cpu()
        label = torch.where(y.T)[1]
        return (pred == label).detach().float().mean().tolist()



"""
This is the implementation of stLearn.
Extended from https://github.com/BiomedicalMachineLearning/stLearn

Reference
----------
Pham, Duy, et al. "stLearn: integrating spatial location, tissue morphology and gene expression to find cell types, cell-cell interactions and spatial trajectories within undissociated tissues." BioRxiv (2020).
"""

# load kmeans from sklearn 
from sklearn.cluster import KMeans

from .louvain import Louvain

# kmeans for adata 
def stKmeans(adata, n_clusters=19,
            init = "k-means++",
            n_init = 10,
            max_iter = 300,
            tol = 1e-4,
            algorithm='auto',
            verbose = False,
            random_state = None,
            use_data='X_pca', key_added="X_pca_kmeans"):
    # kmeans for gene expression data
    kmeans = KMeans(n_clusters=n_clusters, 
                    init=init, 
                    n_init=n_init,
                    max_iter=max_iter,
                    tol=tol,
                    algorithm=algorithm,
                    verbose=verbose,
                    random_state=random_state).fit(adata.obsm[use_data])
    adata.obs[key_added] = kmeans.labels_
    return adata


def stPrepare(adata):
    adata.obs['imagerow'] = adata.obs['x_pixel']
    adata.obs['imagecol'] = adata.obs['y_pixel']
    adata.obs['array_row'] = adata.obs['x']
    adata.obs['array_col'] = adata.obs['y']


class StKmeans:
    """StKmeans class.
    
    Parameters
    ----------
    n_clusters : int optional
        the number of clusters to form as well as the number of centroids to generate.
    init : str optional
        method for initialization: {‘k-means++’, ‘random’}.
    n_init : int optional
        number of time the k-means algorithm will be run with different centroid seeds.
        The final results will be the best output of n_init consecutive runs in terms of inertia.
    max_iter : int optional
        maximum number of iterations of the k-means algorithm for a single run.
    tol : float optional
        relative tolerance with regards to Frobenius norm of the difference in
        the cluster centers of two consecutive iterations to declare convergence.
    algorithm : str optional
        {“lloyd”, “elkan”, “auto”, “full”}, by default "auto"
    verbose : bool optional
        verbosity mode.
    random_state : int optional
        determines random number generation for centroid initialization. 
        Use an int to make the randomness deterministic. 
    use_data : str optional
        by default 'X_pca'.
    key_added : str optional
        by default 'X_pca_kmeans'.
    
    """
    def __init__(self, n_clusters=19,
            init = "k-means++",
            n_init = 10,
            max_iter = 300,
            tol = 1e-4,
            algorithm='auto',
            verbose = False,
            random_state = None,
            use_data='X_pca', key_added="X_pca_kmeans"):
        self.use_data=  use_data
        self.key_added = key_added
        self.model = KMeans(n_clusters=n_clusters,init=init, 
                    n_init=n_init,
                    max_iter=max_iter,
                    tol=tol,
                    algorithm=algorithm,
                    verbose=verbose,
                    random_state=random_state)

    def fit(self, adata):
        """fit function for model training.
        
        Parameters
        ----------
        adata :  
            input data.
        
        Returns
        -------
        None.
        
        """
        self.model.fit(adata.obsm[self.use_data])
        adata.obs[self.key_added] = self.model.labels_

    def predict(self):
        """prediction function.
        
        Parameters
        ----------
     
        Returns
        -------
        self.model.labels_ : 
            predicted label.
        
        """
        self.predict = self.model.labels_
        self.y_pred = self.predict
        return self.predict
    
    def score(self,y_true):
        """score function.
        
        Parameters
        ----------
        adata :  
            input data.
        
        Returns
        -------
        self.score : 
            score.
        
        """
        from sklearn.metrics.cluster import adjusted_rand_score
        score = adjusted_rand_score(y_true, self.y_pred)
        print("ARI {}".format(adjusted_rand_score(y_true, self.y_pred)))
        return score

class StLouvain:
    """StLouvain class.
    """
    def __init__(self):
        self.model = Louvain()

    def fit(self,adata, 
                adj,
                partition=None,
                weight='weight',
                resolution=1.,
                randomize=None,
                random_state=None):
        """fit function for model training.
        
        Parameters
        ----------
        adata :  
            input data.
        adj : 
            adjacent matrix.
        partition : dict optional
            a dictionary where keys are graph nodes and values the part the node
            belongs to
        weight : str, optional
            the key in graph to use as weight. Default to 'weight'
        resolution : float optional
            resolution.
        randomize : boolean, optional
            Will randomize the node evaluation order and the community evaluation
            order to get different partitions at each call
        random_state : int, RandomState instance or None, optional (default=None)
            If int, random_state is the seed used by the random number generator;
            If RandomState instance, random_state is the random number generator;
            If None, the random number generator is the RandomState instance used
            by `np.random`.
        Returns
        -------
        None.
        
        """
        self.data = adata
        self.model.fit(adata, adj, partition, weight, resolution, randomize, random_state)

    def predict(self):
        """prediction function.
        
        Parameters
        ----------
     
        Returns
        -------
        self.y_pred : 
            predicted label.
        
        """
        
        self.y_pred = self.model.predict()
        self.y_pred = [self.y_pred[i] for i in range(len(self.y_pred))]
        self.data.obs['predict'] = self.y_pred
        return self.y_pred 

    def score(self,y_true):
        """score function.
        
        Parameters
        ----------
        adata :  
            input data.
        
        Returns
        -------
        self.score : 
            score.
        
        """
        self.data.obs['ground'] = y_true
        tempdata = self.data.obs.dropna()
        from sklearn.metrics.cluster import adjusted_rand_score
        score = adjusted_rand_score(tempdata['ground'], tempdata['predict'])
        print("ARI {}".format(adjusted_rand_score(tempdata['ground'], tempdata['predict'])))
        return score



import argparse
import os.path as osp

import numpy as np
import pandas as pd

from dance.datasets.singlemodality import CellTypeDataset
from dance.modules.single_modality.cell_type_annotation.actinn import ACTINN


if __name__ == "__main__":
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--batch_size", type=int, default=128, help="Batch size")
    parser.add_argument("--device", default="cpu", help="Computation device.")
    parser.add_argument("--hidden_dims", nargs="+", type=int, default=[100, 50, 25], help="Hidden dimensions.")
    parser.add_argument("--lambd", type=float, default=0.01, help="Regularization parameter")
    parser.add_argument("--learning_rate", type=float, default=0.01, help="Learning rate")
    parser.add_argument("--num_epochs", type=int, default=50, help="Number of epochs")
    parser.add_argument("--print_cost", action="store_true", help="Print cost when training")
    parser.add_argument("--runs", type=int, default=10, help="Number of repetitions")
    parser.add_argument("--seed", type=int, default=0, help="Initial seed random, offset for each repeatition")
    parser.add_argument("--species", default="mouse")
    parser.add_argument("--test_dataset", nargs="+", default=[299], help="List of testing dataset ids.")
    parser.add_argument("--test_dir", default="test")
    parser.add_argument("--tissue", default="Testis")
    parser.add_argument("--train_dataset", nargs="+", default=[2216], help="List of training dataset ids.")
    parser.add_argument("--train_dir", default="train")
    args = parser.parse_args()

    # Load data
    # TODO: make directly pass species, tissue, dataset ids instead of paths, as done in the svm example
    train_data_paths = [
        osp.join(args.train_dir, args.species, f"{args.species}_{args.tissue}{i}_data.csv")
        for i in args.train_dataset
    ]
    train_label_paths = [
        osp.join(args.train_dir, args.species, f"{args.species}_{args.tissue}{i}_celltype.csv")
        for i in args.train_dataset
    ]
    # TODO: multiple test data
    test_data_path = osp.join(args.test_dir, args.species,
                              f"{args.species}_{args.tissue}{args.test_dataset[0]}_data.csv")
    test_label_path = osp.join(args.test_dir, args.species,
                               f"{args.species}_{args.tissue}{args.test_dataset[0]}_celltype.csv")

    dataloader = CellTypeDataset(data_type="actinn", train_set=train_data_paths, train_label=train_label_paths,
                                 test_set=test_data_path, test_label=test_label_path)
    dataloader = dataloader.load_data()
    barcode = dataloader.barcode
    train_set = dataloader.train_set
    train_label = dataloader.train_label
    test_set = dataloader.test_set
    test_label = dataloader.test_label

    # Initialize and train model
    num_genes, num_train_samples = train_set.shape
    num_cell_types = train_label.shape[0]
    print(f"{num_train_samples=:,}, {num_genes=:,}, {num_cell_types=:,}")
    model = ACTINN(num_genes, num_cell_types, hidden_dims=args.hidden_dims, lr=args.learning_rate, device=args.device,
                   num_epochs=args.num_epochs, batch_size=args.batch_size, print_cost=args.print_cost, lambd=args.lambd)

    scores = []
    for k in range(args.runs):
        model.fit(train_set, train_label, seed=args.seed+k)
        test_predict = model.predict(test_set)

        predicted_label = []
        for i in range(len(test_predict)):
            predicted_label.append(dataloader.label_to_type_dict[test_predict[i].item()])
        predicted_label = pd.DataFrame({"cellname": barcode, "celltype": predicted_label})
        # predicted_label.to_csv("predicted_label.txt", sep="\t", index=False)

        scores.append(model.score(test_set, test_label))
        print(f"Run {k + 1:>2d}/{args.runs:>2d}: {scores[-1]}")
    print(f"Score: {np.mean(scores):04.3f} +/- {np.std(scores):04.3f}")

"""To reproduce ACTINN benchmarks, please refer to command lines belows:

Mouse Brain
$ python actinn.py --species mouse --tissue Brain --train_dataset 753 3285 --test_dataset 2695 --lambd 0.1

Mouse Spleen
$ python actinn.py --species mouse --tissue Spleen --train_dataset 1970 --test_dataset 1759 --lambd 0.0001

Mouse Kidney
$ python actinn.py --species mouse --tissue Kidney --train_dataset 4682 --test_dataset 203 --lambd 0.01

"""



import scanpy as sc
import time
import numpy as np
import argparse
import timeit
import time
from dance.modules.single_modality.cell_type_annotation.celltypist import Celltypist
from dance.datasets.singlemodality import CellTypeDataset
import pandas as pd
import ipdb
import anndata as ad
from scipy import sparse


if __name__ == '__main__':
        
    parser = argparse.ArgumentParser()
    parser.add_argument("--random_seed", type=int, default=10)
    parser.add_argument("--train_dataset", default=4682, type=int, help="train id")
    parser.add_argument("--test_dataset", default=203, type=int, help="test id")
    parser.add_argument("--cell_type_train", type=str, help="name for the cell type information for training data", default = "Cell_type")
    parser.add_argument("--cell_type_test", type=str, help="name for the cell type information for test data", default = "Cell_type")
    parser.add_argument("--check_expression", type=bool, help="whether to check the normalization of training and test data", default = False)
    parser.add_argument("--species", default='mouse', type=str)
    parser.add_argument("--tissue", default='Kidney', type=str)
    parser.add_argument("--train_dir", type=str, default='train')
    parser.add_argument("--test_dir", type=str, default='test')
    parser.add_argument("--proj_path", type=str, default='./')
    parser.add_argument("--map_path", type=str, default='map/mouse/')
    parser.add_argument("--n_jobs", type=int, help="Number of jobs", default = 10)
    parser.add_argument("--max_iter", type=int, help="Max iteration during training", default = 5)
    parser.add_argument("--use_SGD", type=bool, help="Training algorithm -- weather it will be stochastic gradient descent", default = True)
    parser.add_argument("--label_conversion", type=bool, help="whether to convert cell type labels between training and test dataset for scoring", default = False)
    args = parser.parse_args()


    dataloader = CellTypeDataset(random_seed = args.random_seed,data_type = "celltypist", proj_path = args.proj_path, train_dir = args.train_dir,test_dir = args.test_dir,train_dataset = args.train_dataset, test_dataset = args.test_dataset, species = args.species, tissue = args.tissue, map_path = args.map_path)
    dataloader.load_data()
    
    model = Celltypist()
    model_fs = model.fit(dataloader.train_adata, labels = args.cell_type_train, check_expression=args.check_expression,n_jobs = args.n_jobs, max_iter = args.max_iter, use_SGD = args.use_SGD)
    predictions = model.predict(dataloader.test_adata,check_expression=args.check_expression)
    accuracy = model.score(dataloader.test_adata, predictions=predictions, labels =args.cell_type_test, map = dataloader.map_dict[args.test_dataset],label_conversion=True)
    print(accuracy)

"""To reproduce CellTypist benchmarks, please refer to command lines belows:

Mouse Brain
$ python celltypist.py --species mouse --tissue Brain --train_dataset 753 --test_dataset 2695 --dLevel Cell_type --dtype Cell_type --label_conversion True

Mouse Spleen
$ python celltypist.py --species mouse --tissue Spleen --train_dataset 1970 --test_dataset 1759 --dLevel Cell_type --dtype Cell_type --label_conversion False

Mouse Kidney
$ python celltypist.py --species mouse --tissue Kidney --train_dataset 4682 --test_dataset 203 --dLevel Cell_type --dtype Cell_type --label_conversion False

"""



# Copyright 2022 DSE lab.  All rights reserved.

# TODO：start from project root directory to solve the package import error
import scanpy as sc
import sys
sys.path.append("../../..")
from dance.datasets.spatial import SpotDataset
from dance.modules.spatial.spatial_domain.stagate import Stagate
from dance.transforms.graph_construct import construct_graph, stagate_construct_graph
from dance.transforms.preprocess import normalize_total, prefilter_cells, prefilter_genes, prefilter_specialgenes, normalize, log1p
import argparse


if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument("--sample_number", type=str, default="151676", \
		help="12 samples of human dorsolateral prefrontal cortex dataset \
		supported in the task of spatial domain task.")
	parser.add_argument("--hidden_dims", type=list, default=[512, 32], help="hidden dimensions")
	parser.add_argument("--rad_cutoff", type=int, default=150, help="")
	parser.add_argument("--n_epochs", type=int, default=1000, help="epochs")
	parser.add_argument("--high_variable_genes", type=int, default=3000, help="")
	args = parser.parse_args()
 

	# from dance.modules.spatial.spatial_domain.stagan import Stagate
	# get data 
	dataset = SpotDataset(args.sample_number, data_dir="../../../data/spot")
	## dataset.data has repeat name , be careful

	# preprocess data 
	dataset.data.var_names_make_unique()

	sc.pp.highly_variable_genes(dataset.data, flavor="seurat_v3", n_top_genes=args.high_variable_genes)
	normalize_total(dataset.data)
	log1p(dataset.data)

	dataset.adj = stagate_construct_graph(dataset.data, rad_cutoff=args.rad_cutoff)


	hidden_dims=args.hidden_dims

	hidden_dims = [args.high_variable_genes] + hidden_dims

	model = Stagate(hidden_dims)

	model.fit(dataset.data, dataset.adj, n_epochs=args.n_epochs)
	predict = model.predict()
	print(model.score())

""" To reproduce Stagate on other samples, please refer to command lines belows:
NOTE: since the stagate method is unstable, you have to run at least 5 times to get 
      best performance. (same with original Stagate paper)

human dorsolateral prefrontal cortex sample 151676:
python stagate.py --sample_number=151676 

human dorsolateral prefrontal cortex sample 151507:
python stagate.py --sample_number=151507 
"""



import argparse
from pprint import pprint

from dance.datasets.spatial import CellTypeDeconvoDatasetLite
from dance.modules.spatial.cell_type_deconvo.card import Card

# TODO: make this a property of the dataset class?
DATASETS = ["CARD_synthetic", "GSE174746", "SPOTLight_synthetic"]

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("--dataset", default="CARD_synthetic", choices=DATASETS, help="Name of the benchmarking dataset.")
parser.add_argument("--datadir", default="data/spatial", help="Directory to save the data.")
parser.add_argument("--max_iter", type=int, default=10, help="Maximum optimization iteration.")
parser.add_argument("--epsilon", type=float, default=1e-10, help="Optimization threshold.")
parser.add_argument("--location_free", action="store_true", help="Do not supply spatial location if set.")
args = parser.parse_args()
pprint(vars(args))

# Load dataset
dataset = CellTypeDeconvoDatasetLite(data_id=args.dataset, data_dir=args.datadir)
sc_count = dataset.data["ref_sc_count"]
sc_meta = dataset.data["ref_sc_annot"]
spatial_count = dataset.data["mix_count"]
true_p = dataset.data["true_p"]
spatial_location = None if args.location_free else dataset.data["spatial_location"]
ct_select = sorted(set(sc_meta.cellType.unique().tolist()) & set(true_p.columns.tolist()))
print(f"{ct_select=}")

# Initialize and train moel
crd = Card(
    sc_count=sc_count,
    sc_meta=sc_meta,
    spatial_count=spatial_count,
    spatial_location=spatial_location,
    ct_varname="cellType",
    ct_select=ct_select,
    cell_varname=None,
    sample_varname=None,
)
crd.fit(max_iter=args.max_iter, epsilon=args.epsilon)

# Evaluate
pred = crd.predict()
rmse = crd.score(pred, true_p[ct_select].values)

print(f"Predicted cell-type proportions of  sample 1: {pred[0].round(3)}")
print(f"True cell-type proportions of  sample 1: {true_p[ct_select].iloc[0].tolist()}")
print(f"RMSE = {rmse:6.3f}")

"""To reproduce CARD benchmarks, please refer to command lines belows:

CARD synthetic
$ python card.py --dataset CARD_synthetic

GSE174746
$ python card.py --dataset GSE174746 --location_free

SPOTLight synthetic
$ python card.py --dataset SPOTLight_synthetic --location_free

"""
